# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module responses provides helper functions for different response types."""

from functools import partial
from json import dumps as json_dumps

from aiohttp import web


def document_response(document=None,
                      status=200,
                      reason=None,
                      headers=None,
                      content_type='application/vnd.api+json',
                      dumps=partial(json_dumps, default=str)):
    """Creates a response from a document."""

    # pylint: disable=too-many-arguments

    if headers is None:
        headers = {}
    return web.Response(body=dumps(document),
                        status=status,
                        reason=reason,
                        headers=headers,
                        content_type=content_type)


def no_data_response(status=204,
                     reason=None,
                     headers=None,
                     content_type='application/vnd.api+json'):
    """Create a 204 NO DATA response."""

    if headers is None:
        headers = {}
    return web.Response(body='',
                        status=status,
                        reason=reason,
                        headers=headers,
                        content_type=content_type)


def errors_response(errors,
                    reason=None,
                    headers=None,
                    content_type='application/vnd.api+json',
                    dumps=partial(json_dumps, default=str)):
    """Creates a response from errors."""

    statuses = {int(error.status) for error in errors}
    if len(statuses) == 1:
        status = next(iter(statuses))
    else:
        status_classes = {status // 100 for status in statuses}
        status = min(status_classes) * 100
    document = {'errors': [error.to_error() for error in errors]}
    return web.Response(body=dumps(document),
                        status=status,
                        reason=reason,
                        headers=headers,
                        content_type=content_type)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Fixtures for dealing with SQL databases."""

import subprocess
from os import environ
from re import search as re_search
from time import sleep

import pytest
from docker import from_env as docker_from_env

from ajsonapi.application import Application

NETWORK_NAME = 'test-network'
CONTAINER_NAME = 'test-container'

POSTGRES_USER = 'user'
POSTGRES_PASSWORD = 'password'
POSTGRES_PORT = 9900
POSTGRES_DB = 'db'
POSTGRES_LOCAL_DIR = '/tmp/pgdata/ajsonapi'

if 'PYTEST_XDIST_WORKER' in environ:  # pragma: no cover
    WORKER = environ['PYTEST_XDIST_WORKER']
    NETWORK_NAME += f'-{WORKER}'
    CONTAINER_NAME += f'-{WORKER}'
    POSTGRES_PORT += int(re_search(r'\d+$', WORKER).group())
    POSTGRES_LOCAL_DIR += f'_{WORKER}'
    del WORKER

docker = docker_from_env(version='auto')  # pylint: disable=invalid-name


def create_postgres_network(name):
    """Create a docker Postgres network for the functional tests."""
    return docker.networks.create(name, driver='bridge')


def create_postgres_container(name, network):
    """Create a docker Postgres container for the functional tests."""
    kwargs = {
        'name': name,
        'environment': {
            'POSTGRES_USER': POSTGRES_USER,
            'POSTGRES_PASSWORD': POSTGRES_PASSWORD,
            'POSTGRES_DB': POSTGRES_DB,
        },
        'volumes': {
            POSTGRES_LOCAL_DIR: {
                'bind': '/var/lib/postgresql/data',
                'mode': 'rw',
            },
        },
        'network': network.name,
        'ports': {
            '5432/tcp': POSTGRES_PORT,
        },
        'detach': True,
        'ipc_mode': 'shareable',
    }
    return docker.containers.create('postgres', **kwargs)


@pytest.fixture(scope='session')
def postgres_container():
    """Pytest fixture for creating a Docker Postgres container."""

    network = create_postgres_network(NETWORK_NAME)
    container = create_postgres_container(CONTAINER_NAME, network)
    container.start()

    proc = subprocess.run('docker network ls --format "{{.Name}}"',
                          shell=True,
                          check=True,
                          capture_output=True)
    names = ' '.join(proc.stdout.decode().strip().split('\n'))
    proc = subprocess.run(f'docker network inspect {names}',
                          shell=True,
                          check=False,
                          capture_output=True)
    print(proc.stdout.decode().strip())
    proc = subprocess.run(f'docker container inspect {CONTAINER_NAME}',
                          shell=True,
                          check=True,
                          capture_output=True)
    print(proc.stdout.decode().strip())

    yield container
    container.stop()
    container.remove()
    network.remove()


@pytest.fixture()
def postgres_database(request, postgres_container):
    """Pytest fixture for creating a database in a Docker Postgres container.
    """
    # pylint: disable=redefined-outer-name

    # pylint: disable=protected-access
    database_name = request._pyfuncitem.name
    if database_name.startswith('test_'):
        database_name = database_name[5:]
    if database_name.endswith('[pyloop]'):
        database_name = database_name[:-8]
    kwargs = {
        'stdin': True,
        'tty': True,
    }

    result = postgres_container.exec_run(
        f'psql -U {POSTGRES_USER} -d {POSTGRES_DB} '
        f'-c "CREATE DATABASE {database_name}"', **kwargs)
    okay = (result.exit_code == 0)
    while not okay:  # pragma: no cover
        sleep(.1)
        result = postgres_container.exec_run(
            f'psql -U {POSTGRES_USER} -d {POSTGRES_DB} '
            f'-c "CREATE DATABASE {database_name}"', **kwargs)
        okay = (result.exit_code == 0)
    # Gitlab CI/CD runs tests within a container so we will have to find
    # the internal Postgres URI (instead of the external one
    # (localhost:{POSTGRES_PORT}) for the development machine).
    proc = subprocess.run(
        'docker inspect --format '
        "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}' "
        f'{CONTAINER_NAME}',
        shell=True,
        check=True,
        capture_output=True)
    internal_ip_address = proc.stdout.decode().strip()
    yield (f'postgresql://{POSTGRES_USER}:{POSTGRES_PASSWORD}@'
           f'{internal_ip_address}:5432/{database_name}')
    postgres_container.exec_run((f'psql -U {POSTGRES_USER} -d {POSTGRES_DB} '
                                 f'-c "DROP DATABASE {database_name}"'),
                                **kwargs)


@pytest.fixture()
async def app(postgres_database):
    """Pytest fixture for creating the JSON API application."""
    # pylint: disable=redefined-outer-name

    app = Application()
    app.verify_model()
    await app.connect_database(postgres_database)
    await app.create_tables()
    app.add_event_socket_route()
    app.add_json_api_routes()
    yield app
    await app.disconnect_database()


@pytest.fixture()
def client(loop, app, aiohttp_client):
    """Pytest fixture for creating a JSON API test client."""
    # pylint: disable=redefined-outer-name

    return loop.run_until_complete(aiohttp_client(app.app))



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module uri.relationship provides things related to relationship URIs."""

from abc import abstractmethod
from asyncio import Lock, create_task, gather
from itertools import chain
from re import compile as re_compile

from asyncpg.exceptions import ForeignKeyViolationError, UniqueViolationError

from ajsonapi.conversions import id_number_to_name
from ajsonapi.document import (
    ResourceIdentifierObjectDocument,
    ResourceIdentifierObjectsDocument,
    parse_request_document_data,
)
from ajsonapi.environments import GetObjectIncludedEnvironment
from ajsonapi.errors import (
    DocumentDataNonexistentIdError,
    DocumentDataRelationshipConflictError,
    MethodNotAllowedError,
    ResourceNotFoundError,
    ResourceNotFoundMalformedIdError,
)
from ajsonapi.exceptions import ErrorsException
from ajsonapi.id_value import IdValue
from ajsonapi.logging import log
from ajsonapi.responses import (
    document_response,
    errors_response,
    no_data_response,
)
from ajsonapi.uri.collection import Collection
from ajsonapi.uri.resource import Resource

RE_FOREIGN_KEY_VIOLATION_ERROR_DETAIL = re_compile(
    r'Key \(([^\)]*)\)=\(([^\)]*)\) is not present in table "([^"]*)".')


class Relationship(Resource):
    """Class Relationship represents the relationship URI
    (/{collection}/{id}/relationships/{relationship}).
    """

    def __init__(self, collection, id_, relationship):
        # pylint: disable=too-many-arguments
        self.collection = collection
        self.id = id_  # pylint: disable=invalid-name
        self.relationship = relationship

        self.table = collection.table

    @abstractmethod
    async def parse_document(self, request):
        """Parses the request document for a valid resource object.

        Args:
            request (Request): the request to be parsed.

        Returns:
            A ResourceIdentifierObjectDocument or a
            ResourceIdentifierObjectsDocument (plural).

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

    async def get(self, query):
        """Produces the response for the GET
        /{collection}/{id}/relationships/{relationship} request.
        """
        links = {
            'self':
                self.path(),
            'related': (f'/{self.collection.name}/{self.id.name}'
                        f'/{self.relationship.name}')
        }
        return document_response(await self.to_document(query=query,
                                                        links=links))

    async def to_document(self, query=None, links=None):
        """Creates the document for the resource objects in this resource
        collection.

        Args:
            links (dict): value for the links field in the response.
        """

        data, included = await self.to_data_included(query)
        document = {'data': data}
        if included is not None:
            document['included'] = included
        if links:
            document['links'] = links
        return document

    async def to_data_included(self, query):
        """Creates the document's data and included member values for this
        collection.
        """
        env = GetRelationshipEnvironment(self, query)
        await env.run()
        return env.data, env.included

    @abstractmethod
    async def get_relationship_data(self, env):
        """Creates the response data member for GET /{collection} responses.

        Specifically, after this function call completes, the env's data
        member contains something like the following object:

        {
            'type': 'centers',
            'id': '1'
        }
        """

    @abstractmethod
    async def patch(self, document, query):
        """Produces the response for the PATCH
        /{collection}/{id}/relationships/{relationship} request.
        """

    @abstractmethod
    async def post(self, document, query):
        """Produces the response for a POST
        /{collection}/{id}/relationships/{relationship} request.
        """

    @abstractmethod
    async def delete(self, document, query):
        """Produces the response for a DELETE
        /{collection}/{id}/relationships/{relationship} request.
        """

    @abstractmethod
    def put(self):
        """Produces the response for a PUT
        /{collection}/{id}/relationships/{relationship} request.
        """

    async def verify_id_exists(self, connection):
        """Raises an exception in case the id in the URL does not exist."""
        stmt = f'SELECT FROM {self.table.name} WHERE id=$1'
        record = await connection.fetchrow(stmt, self.id.number)
        if record is None:
            raise ErrorsException([
                ResourceNotFoundError(f'/{self.collection.name}/{self.id.name}')
            ])

    def path(self):
        """Creates the path to the relationship."""
        return (f'/{self.collection.name}/{self.id.name}/'
                f'relationships/{self.relationship.name}')


class ToOneRelationship(Relationship):
    """Class ToOneRelationship represents the to-one relationship URI
    (/{collection}/{id}/relationships/{relationship}).
    """

    async def parse_document(self, request):
        """Parses the request document for a valid resource object.

        Args:
            request (Request): the request to be parsed.

        Returns:
            A ResourceIdentifierObjectDocument.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

        data = await parse_request_document_data(request)
        document = ResourceIdentifierObjectDocument(data)
        document.parse(self.relationship.rtable.collection.name)
        return document

    async def patch(self, document, query):
        if document.data is not None:
            data_id = document.id
        else:
            data_id = None
        await self.patch_execute(data_id)
        return no_data_response()

    @abstractmethod
    async def patch_execute(self, data_id):
        """Updates the database in response to a PATCH
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-one relationship.
        """

    async def post(self, document, query):
        """Produces the response for a POST
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-one relationship.
        """
        return errors_response([MethodNotAllowedError()],
                               headers={'Allow': 'GET,HEAD,PATCH'})

    async def delete(self, document, query):
        """Produces the response for a DELETE
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-one relationship.
        """
        return errors_response([MethodNotAllowedError()],
                               headers={'Allow': 'GET,HEAD,PATCH'})

    def put(self):
        """Produces the response for a DELETE
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-one relationship.
        """
        return errors_response([MethodNotAllowedError()],
                               headers={'Allow': 'GET,HEAD,PATCH'})


class OneToOneLocalRelationship(ToOneRelationship):
    """Class OneToOneLocalRelationship represents the URI
    (/{collection}/{id}/relationships/{relationship}) for a one-to-one
    relationship with a local foreign id.
    """

    async def get_relationship_data(self, env):
        stmt = (f'SELECT {self.relationship.lfkey} '
                f'FROM {self.table.name} '
                f'WHERE id=$1')
        async with self.table.pool.acquire() as connection:
            log.debug(f'{stmt}: {self.id.number}')
            record = await connection.fetchrow(stmt, self.id.number)
        if record is None:
            raise ErrorsException([
                ResourceNotFoundError(f'/{self.collection.name}/{self.id.name}')
            ])
        record_id = record[0]
        if record_id is None:
            env.data = None
            return
        env.data = {
            'type': self.relationship.rtable.collection.name,
            'id': id_number_to_name(record_id),
        }

    async def patch_execute(self, data_id):
        # Need to remove other lfkeys that point to data_id_number!
        stmt = (f'UPDATE {self.relationship.table.name} '
                f'SET {self.relationship.lfkey} = $1 '
                f'WHERE id=$2')
        if data_id is None:
            data_id_number = None
        else:
            data_id_number = data_id.number
        try:
            async with self.table.pool.acquire() as connection:
                log.debug(f'{stmt}: {data_id_number}, {self.id.number}')
                result = await connection.execute(stmt, data_id_number,
                                                  self.id.number)
            if result == 'UPDATE 0':
                raise ErrorsException([
                    ResourceNotFoundError(
                        f'/{self.collection.name}/{self.id.name}')
                ])
        except UniqueViolationError:
            raise ErrorsException([DocumentDataRelationshipConflictError()])
        except ForeignKeyViolationError:
            raise ErrorsException(
                [DocumentDataNonexistentIdError(f'/data/id/{data_id.name}')])


class OneToOneRemoteRelationship(ToOneRelationship):
    """Class OneToOneRemoteRelationship represents the URI
    (/{collection}/{id}/relationships/{relationship}) for a one-to-one
    relationship with a remote foreign id.
    """

    async def get_relationship_data(self, env):
        stmt = (f'SELECT id '
                f'FROM {self.relationship.rtable.name} '
                f'WHERE {self.relationship.rfkey}=$1')
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                log.debug(f'{stmt}: {self.id.number}')
                record = await connection.fetchrow(stmt, self.id.number)
                if record is None:
                    await self.verify_id_exists(connection)
                    env.data = None
                    return
        env.data = {
            'type': self.relationship.rtable.collection.name,
            'id': id_number_to_name(record[0]),
        }

    async def patch_execute(self, data_id):
        stmt_remove = (f'UPDATE {self.relationship.rtable.name} '
                       f'SET {self.relationship.rfkey} = NULL '
                       f'WHERE {self.relationship.rfkey}=$1')
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                log.debug(f'{stmt_remove}: {self.id.number}')
                await connection.execute(stmt_remove, self.id.number)
                if data_id:
                    stmt_add = (f'UPDATE {self.relationship.rtable.name} '
                                f'SET {self.relationship.rfkey} = $1 '
                                f'WHERE id=$2 AND '
                                f'({self.relationship.rfkey} IS NULL OR '
                                f'{self.relationship.rfkey}=$1)')
                    try:
                        log.debug(f'{stmt_add}: {self.id.number}, '
                                  f'{data_id.number}')
                        result = await connection.execute(
                            stmt_add, self.id.number, data_id.number)
                        if result == 'UPDATE 0':
                            stmt_verify = (f'SELECT '
                                           f'FROM'
                                           f' {self.relationship.rtable.name} '
                                           f'WHERE id=$1')
                            log.debug(stmt_verify)
                            records = await connection.fetch(
                                stmt_verify, data_id.number)
                            if records == []:
                                raise ErrorsException([
                                    DocumentDataNonexistentIdError(
                                        f'/data/id/{data_id.name}')
                                ])
                            raise ErrorsException(
                                [DocumentDataRelationshipConflictError()])
                    except ForeignKeyViolationError:
                        raise ErrorsException([
                            ResourceNotFoundError(
                                f'/{self.collection.name}/{self.id.name}')
                        ])


class ManyToOneRelationship(ToOneRelationship):
    """Class ManyToOneRelationship represents the URI many-to-one relationship
    URI (/{collection}/{id}/relationships/{relationship}).
    """

    async def get_relationship_data(self, env):
        stmt = (f'SELECT {self.relationship.lfkey} '
                f'FROM {self.table.name} '
                f'WHERE id=$1')
        async with self.table.pool.acquire() as connection:
            log.debug(f'{stmt}: {self.id.number}')
            record = await connection.fetchrow(stmt, self.id.number)
        if record is None:
            raise ErrorsException([
                ResourceNotFoundError(f'/{self.collection.name}/{self.id.name}')
            ])
        record_id = record[0]
        if record_id is None:
            env.data = None
            return
        env.data = {
            'type': self.relationship.rtable.collection.name,
            'id': id_number_to_name(record_id),
        }

    async def patch_execute(self, data_id):
        stmt = (f'UPDATE {self.table.name} '
                f'SET {self.relationship.lfkey} = $1 '
                f'WHERE id=$2')
        if data_id is None:
            data_id_number = None
        else:
            data_id_number = data_id.number
        try:
            async with self.table.pool.acquire() as connection:
                log.debug(f'{stmt}: {data_id_number}, {self.id.number}')
                result = await connection.execute(stmt, data_id_number,
                                                  self.id.number)
            if result == 'UPDATE 0':
                raise ErrorsException([
                    ResourceNotFoundError(
                        f'/{self.collection.name}/{self.id.name}')
                ])
        except ForeignKeyViolationError:
            raise ErrorsException(
                [DocumentDataNonexistentIdError(f'/data/id/{data_id.name}')])


class ToManyRelationship(Relationship):
    """Class ToOneRelationship represents the to-many relationship URI
    (/{collection}/{id}/relationships/{relationship}).
    """

    async def parse_document(self, request):
        """Parses the request document for a valid resource object.

        Args:
            request (Request): the request to be parsed.

        Returns:
            A ResourceIdentifierObjectsDocument.

        Raises:
            ErrorsException with all the errors in case the document does not
            represent a valid resource object.
        """

        data = await parse_request_document_data(request)
        document = ResourceIdentifierObjectsDocument(data)
        document.parse(self.relationship.rtable.collection.name)
        return document

    async def patch(self, document, query):
        await self.patch_execute(document.ids)
        return no_data_response()

    @abstractmethod
    async def patch_execute(self, data_ids):
        """Updates the database in response to a PATCH
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-many relationship.
        """

    async def post(self, document, query):
        """Produces the response for a POST
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-many relationship.
        """
        await self.post_execute(document.ids)
        return no_data_response()

    @abstractmethod
    async def post_execute(self, data_ids):
        """Updates the database in response to a POST
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-many relationship.
        """

    async def delete(self, document, query):
        """Produces the response for a POST
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-many relationship.
        """
        await self.delete_execute(document.ids)
        return no_data_response()

    def put(self):
        """Produces the response for a DELETE
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-one relationship.
        """
        return errors_response([MethodNotAllowedError()],
                               headers={'Allow': 'DELETE,GET,HEAD,PATCH,POST'})

    @abstractmethod
    async def delete_execute(self, data_ids):
        """Updates the database in response to a POST
        /{collection}/{id}/relationships/{relationship} request where
        {relationship} is a to-many relationship.
        """


class OneToManyRelationship(ToManyRelationship):
    """Class OneToManyRelationship represents the URI one-to-many relationship
    URI (/{collection}/{id}/relationships/{relationship}).
    """

    async def get_relationship_data(self, env):
        stmt = (f'SELECT id '
                f'FROM {self.relationship.rtable.name} '
                f'WHERE {self.relationship.rfkey}=$1')
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                log.debug(f'{stmt}: {self.id.number}')
                records = await connection.fetch(stmt, self.id.number)
                if records == []:
                    await self.verify_id_exists(connection)
                    env.data = []
                    return
        env.data = [{
            'type': self.relationship.rtable.collection.name,
            'id': id_number_to_name(record[0]),
        } for record in records]

    async def patch_execute(self, data_ids):
        stmt_remove = (f'UPDATE {self.relationship.rtable.name} '
                       f'SET {self.relationship.rfkey} = NULL '
                       f'WHERE {self.relationship.rfkey}=$1')
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                log.debug(f'{stmt_remove}: {self.id.number}')
                await connection.execute(stmt_remove, self.id.number)
                if data_ids:
                    stmt_add = (f'UPDATE {self.relationship.rtable.name} '
                                f'SET {self.relationship.rfkey} = $1 '
                                f'WHERE id=any($2::UUID[])')
                    data_id_numbers = [data_id.number for data_id in data_ids]
                    try:
                        log.debug(f'{stmt_add}: {self.id.number}, '
                                  f'{data_id_numbers}')
                        result = await connection.execute(
                            stmt_add, self.id.number, data_id_numbers)
                        if result != f'UPDATE {len(data_ids)}':
                            stmt_verify = (f'SELECT id '
                                           f'FROM'
                                           f' {self.relationship.rtable.name} '
                                           f'WHERE id=ANY($1::UUID[])')
                            log.debug(stmt_verify)
                            records = await connection.fetch(
                                stmt_verify, data_id_numbers)
                            records_set = set(records)
                            errors = [
                                DocumentDataNonexistentIdError(
                                    f'/data/{index}/id/{data_id.name}')
                                for index, data_id in enumerate(data_ids)
                                if data_id.number not in records_set
                            ]
                            raise ErrorsException(errors)
                    except ForeignKeyViolationError:
                        raise ErrorsException([
                            ResourceNotFoundError(
                                f'/{self.collection.name}/{self.id.name}')
                        ])
                else:
                    await self.verify_id_exists(connection)

    async def post_execute(self, data_ids):
        if data_ids:
            stmt = (f'UPDATE {self.relationship.rtable.name} '
                    f'SET {self.relationship.rfkey} = $1 '
                    f'WHERE id=ANY($2::UUID[])')
            data_id_numbers = [data_id.number for data_id in data_ids]
            async with self.table.pool.acquire() as connection:
                async with connection.transaction():
                    try:
                        log.debug(f'{stmt}: {self.id.number}, '
                                  f'{data_id_numbers}')
                        result = await connection.execute(
                            stmt, self.id.number, data_id_numbers)
                    except ForeignKeyViolationError:
                        raise ErrorsException([
                            ResourceNotFoundError(
                                f'/{self.collection.name}/{self.id.name}')
                        ])
                    if result != f'UPDATE {len(data_ids)}':
                        stmt_verify = (f'SELECT id '
                                       f'FROM {self.relationship.rtable.name} '
                                       f'WHERE id=ANY($1::UUID[])')
                        log.debug(f'{stmt_verify}: {data_id_numbers}')
                        records = await connection.fetch(
                            stmt_verify, data_id_numbers)
                        records_set = set(records)
                        errors = [
                            DocumentDataNonexistentIdError(
                                f'/data/{index}/id/{data_id.name}')
                            for index, data_id in enumerate(data_ids)
                            if data_id.number not in records_set
                        ]
                        raise ErrorsException(errors)
        else:
            async with self.table.pool.acquire() as connection:
                await self.verify_id_exists(connection)

    async def delete_execute(self, data_ids):
        if data_ids:
            stmt = (f'UPDATE {self.relationship.rtable.name} '
                    f'SET {self.relationship.rfkey} = NULL '
                    f'WHERE {self.relationship.rfkey}=$1 AND '
                    f'id=ANY($2::UUID[])')
            data_id_numbers = [data_id.number for data_id in data_ids]
            async with self.table.pool.acquire() as connection:
                async with connection.transaction():
                    log.debug(f'{stmt}: {self.id.number}, {data_id_numbers}')
                    result = await connection.execute(stmt, self.id.number,
                                                      data_id_numbers)
                    if result == 'UPDATE 0':
                        await self.verify_id_exists(connection)
        else:
            async with self.table.pool.acquire() as connection:
                await self.verify_id_exists(connection)


class ManyToManyRelationship(ToManyRelationship):
    """Class ManyToManyRelationship represents the URI many-to-many relationship
    URI (/{collection}/{id}/relationships/{relationship}).
    """

    async def get_relationship_data(self, env):
        stmt = (f'SELECT {self.relationship.rafkey} '
                f'FROM {self.relationship.atable.name} '
                f'WHERE {self.relationship.lafkey}=$1')
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                log.debug(f'{stmt}: {self.id.number}')
                records = await connection.fetch(stmt, self.id.number)
                if records == []:
                    await self.verify_id_exists(connection)
                    env.data = []
                    return
        env.data = [{
            'type': self.relationship.rtable.collection.name,
            'id': id_number_to_name(record[0]),
        } for record in records]

    async def patch_execute(self, data_ids):
        stmt_remove = (f'DELETE FROM {self.relationship.atable.name} '
                       f'WHERE {self.relationship.lafkey}=$1')
        stmt_add = (f'INSERT INTO {self.relationship.atable.name} '
                    f'({self.relationship.lafkey}, '
                    f'{self.relationship.rafkey}) '
                    f'VALUES ($1, $2)')
        values = [(self.id.number, data_id.number) for data_id in data_ids]
        async with self.table.pool.acquire() as connection:
            async with connection.transaction():
                try:
                    async with connection.transaction():
                        log.debug(f'{stmt_remove}: {self.id.number}')
                        await connection.execute(stmt_remove, self.id.number)
                        log.debug(f'{stmt_add}: {values}')
                        await connection.executemany(stmt_add, values)
                except ForeignKeyViolationError as exc:
                    async with connection.transaction():
                        await self.on_foreign_key_violation_error(
                            connection, exc, data_ids)

    async def post_execute(self, data_ids):
        if data_ids:
            # To prevent duplicates, we first remove any existing associations
            stmt_remove = (f'DELETE FROM {self.relationship.atable.name} '
                           f'WHERE {self.relationship.lafkey}=$1 AND '
                           f'{self.relationship.rafkey}=ANY($2::UUID[])')
            stmt_add = (f'INSERT INTO {self.relationship.atable.name} '
                        f'({self.relationship.lafkey}, '
                        f'{self.relationship.rafkey}) '
                        f'VALUES ($1, $2)')
            data_id_numbers = [data_id.number for data_id in data_ids]
            values = [(self.id.number, data_id.number) for data_id in data_ids]
            async with self.table.pool.acquire() as connection:
                async with connection.transaction():
                    try:
                        async with connection.transaction():
                            log.debug(f'{stmt_remove}: {self.id.number}, '
                                      f'{data_id_numbers}')
                            await connection.execute(stmt_remove,
                                                     self.id.number,
                                                     data_id_numbers)
                            log.debug(f'{stmt_add}: {values}')
                            await connection.executemany(stmt_add, values)
                    except ForeignKeyViolationError as exc:
                        async with connection.transaction():
                            await self.on_foreign_key_violation_error(
                                connection, exc, data_ids)
        else:
            async with self.table.pool.acquire() as connection:
                await self.verify_id_exists(connection)

    async def delete_execute(self, data_ids):
        if data_ids:
            stmt = (f'DELETE FROM {self.relationship.atable.name} '
                    f'WHERE {self.relationship.lafkey}=$1 AND '
                    f'{self.relationship.rafkey}=ANY($2::UUID[])')
            data_id_numbers = [data_id.number for data_id in data_ids]
            async with self.table.pool.acquire() as connection:
                async with connection.transaction():
                    log.debug(f'{stmt}: {self.id.number}, {data_id_numbers}')
                    result = await connection.execute(stmt, self.id.number,
                                                      data_id_numbers)
                    if result == 'DELETE 0':
                        await self.verify_id_exists(connection)
        else:
            async with self.table.pool.acquire() as connection:
                await self.verify_id_exists(connection)

    async def on_foreign_key_violation_error(self, connection, exc, data_ids):
        """Handles a ForeignKeyViolationError on changing the association
        table.
        """
        detail = exc.detail  # pylint: disable=no-member
        groups = RE_FOREIGN_KEY_VIOLATION_ERROR_DETAIL.match(detail).groups()
        if groups[0] == self.relationship.lafkey:
            raise ErrorsException([
                ResourceNotFoundError(f'/{self.collection.name}/{self.id.name}')
            ])
        # groups[0] == self.relationship.rafkey

        stmt_verify = (f'SELECT {self.relationship.rafkey} '
                       f'FROM {self.relationship.atable.name} '
                       f'WHERE {self.relationship.lafkey}=$1')
        log.debug(f'{stmt_verify}: {self.id.number}')
        records = await connection.fetch(stmt_verify, self.id.number)
        errors = [
            DocumentDataNonexistentIdError(f'/data/{index}/id/{data_id.name}')
            for index, data_id in enumerate(data_ids)
            if data_id.number not in set(records)
        ]
        raise ErrorsException(errors)


class GetRelationshipEnvironment:
    """Class GetRelatedResourceEnvironment is the execution environment
    for a generating a GET /{collection}/{id}/relationships/{relationship}
    response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, relationship, query):
        self.relationship = relationship
        self.query = query
        self.tasks = []
        self.connection = None
        self.lock = Lock()
        self.data_environment = None
        self.data = None
        self.included_environments = []
        self.included = None

    def create_data_environment(self):
        """Creates the environment for collecting the data member of a GET
        /{collection}/{id}/relationships/{relationship} response.
        """
        self.data_environment = GetRelationshipDataEnvironment(self)

    def create_included_environments(self, ids, includes):
        """Creates the environment for collecting the included member of a
        GET /{collection}/{id}/relationships/{relationships} response.
        """
        if includes is None:
            return
        for relationship, remote_includes in includes.items():
            remote_collection = relationship.rtable.collection
            remote_ids = relationship.object_remote_ids(ids)
            self.included_environments.append(
                GetObjectIncludedEnvironment(self.relationship.id.number,
                                             remote_collection, remote_ids,
                                             self))
            if remote_includes:
                self.create_included_environments(remote_ids, remote_includes)

    def create_tasks(self):
        """Create all tasks for querying the database to create a GET
        /{collection}/{id}/relationships/{relationship} response.
        """
        self.create_data_environment()
        self.create_included_environments(None, self.query['includes'])
        self.tasks = [
            create_task(env.run()) for env in chain([self.data_environment],
                                                    self.included_environments)
        ]

    async def run(self):
        """Executes the GET /{collection}/{id}/relationships/{relationship}
        tasks.
        """

        async with self.relationship.table.pool.acquire() as self.connection:
            async with self.connection.transaction():
                self.create_tasks()
                await gather(*self.tasks)
        self.data = self.data_environment.data
        if self.query['includes']:
            self.included = list(
                chain.from_iterable(env.included_by_id.values()
                                    for env in self.included_environments))


class GetRelationshipDataEnvironment:
    """Class GetRelationshipDataEnvironment is the execution environment for
    the data part of the GET /{collection}/{id}/relationships/{relationship}
    response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, env):
        self.relationship = env.relationship
        self.query = env.query
        self.tasks = []
        self.connection = env.connection
        self.lock = env.lock
        self.data = None

    def create_tasks(self):
        """Create all tasks for querying the database to create a GET
        /{collection}/{id}/relationships/{relationship} response.
        """
        self.tasks = [
            create_task(self.relationship.get_relationship_data(self))
        ]

    async def run(self):
        """Executes the environment to generate the data member of the GET
        /{collection}/{id}/relationships/{relationship} response.
        """
        self.create_tasks()
        await gather(*self.tasks)


def make_relationship(collection_name, id_name, relationship_name):
    """Returns the relationship URI
    (/{collection}/{id}/relationships/{relationship}) corresponding to the user
    provided model relationships.

    Args:
        collection_name (str): Name of the collection.
        id_name (str): Name of the id.
        relationship_name (str): Name of the relationship.

    Returns:
        An instantiated object of the correct derived class of Relationship.
    """

    try:
        collection = Collection.by_name[collection_name]
    except KeyError:
        raise ErrorsException([ResourceNotFoundError(f'/{collection_name}')])

    try:
        id_ = IdValue(id_name)
    except ValueError:
        raise ErrorsException(
            [ResourceNotFoundMalformedIdError(f'/{collection_name}/{id_name}')])

    try:
        relationship = getattr(collection.table.___, relationship_name)
    except AttributeError:
        raise ErrorsException([
            ResourceNotFoundError(f'/{collection_name}/{id_name}/'
                                  f'relationships/{relationship_name}')
        ])

    return relationship.make_uri_relationship(collection, id_)


def parse(request):
    """Gets a relationship URI associated with a request.

    Args:
        request: Incoming Http(s) request.

    Returns:
        A URI relationship.

    Exceptions:
        ErrorsException: Exception containing a 'resource not found' error.
    """

    collection_name = request.match_info['collection']
    id_name = request.match_info['id']
    relationship_name = request.match_info['relationship']
    return make_relationship(collection_name, id_name, relationship_name)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module headers provides parsing of JSON API headers"""

from ajsonapi.errors import NotAcceptableError, UnsupportedMediaTypeError
from ajsonapi.exceptions import ErrorsException


def parse(request):
    """Parses the headers for errors.

    Args:
        request: Request whose headers are parsed.

    Returns:
        The request's headers.

    Raises:
        An ErrorsException containing the found header errors.
    """

    errors = []
    headers = request.headers

    # [https://jsonapi.org/format/#content-negotiation-servers]
    # Servers MUST respond with a '415 Unsupported Media Type' status code if
    # a request specifies the header 'Content-Type: application/vnd.api+json'
    # with any media type parameters.
    try:
        content_type = headers['Content-Type']
        media_type, *parameters = content_type.split(';')
        if media_type != 'application/vnd.api+json' or parameters:
            errors.append(UnsupportedMediaTypeError())
    except KeyError:
        pass

    # [https://jsonapi.org/format/#content-negotiation-servers]
    # Servers MUST respond with a '406 Not Acceptable' status code if a
    # request's 'Accept' header contains the JSON API media type and all
    # instances of that media type are modified with media type parameters.
    accept = headers['Accept']  # aiohttp provides default Accept header
    json_api_params = []
    accept_values = accept.split(',')
    for accept_value in accept_values:
        media_range, *accept_params = accept_value.split(';')
        if media_range == 'application/vnd.api+json':
            json_api_params.append(accept_params)
    if json_api_params != [] and [] not in json_api_params:
        errors.append(NotAcceptableError())

    if errors:
        raise ErrorsException(errors)
    return headers



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module attribute provides class Attribute."""

from csv import reader as csv_reader

from ajsonapi.field import Field


class Attribute(Field):
    """Attributes are used to specify data members of object model classes
    (classes derived from JSON_API).
    """

    def __init__(self, type_):
        self.type_ = type_
        self.name = ''  # Overridden in JSON_API.__init_subclass__

    def filter_condition(self, values):
        sql_values = [
            self.type_.sql(value) for value in next(csv_reader([values]))
        ]
        return f"{self.name} IN ({','.join(sql_values)})"

    def sql(self):
        """Creates the SQL column definition for this attribute.

        Returns:
            A string containing the SQL column definition for this attribute.
        """

        return f'{self.name} {self.type_.name}'

    def __str__(self):
        return self.name



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module errors provides errors related to request parsing."""

from abc import ABC, abstractmethod


class Error(ABC):
    """Abstract base class for all JSON API errors."""

    # pylint: disable=too-few-public-methods

    @abstractmethod
    def to_error(self):
        """Returns a JSON API error document for use in the response
        document's errors array.
        """


class MethodNotAllowedError(Error):
    """Method not allowed for URL."""

    # pylint: disable=too-few-public-methods

    status = 405

    def to_error(self):
        return {
            'status': '405',
            'title': 'Method not allowed.',
        }


class ResourceNotFoundError(Error):
    """Errors related to an invalid url in a request."""

    # pylint: disable=too-few-public-methods

    status = 404

    def __init__(self, path, about=None):
        self.path = path
        self.about = about

    def to_error(self):
        return {
            'status': '404',
            'source': {
                'path': self.path,
            },
            'title': 'Resource does not exist.',
        }


class ResourceNotFoundMalformedIdError(Error):
    """Errors related to an invalid url due to a malformed id in a request."""

    # pylint: disable=too-few-public-methods

    status = 404

    def __init__(self, path):
        self.path = path

    def to_error(self):
        return {
            'status': '404',
            'source': {
                'path': self.path,
            },
            'title': 'Resource does not exist.',
            'detail': 'Malformed id.',
        }


class QueryParameterUnsupportedError(Error):
    """Errors related to unsupported query parameters in a request."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, parameter):
        self.parameter = parameter

    def to_error(self):
        return {
            'status': '400',
            'source': {
                'parameter': self.parameter,
            },
            'title': 'Parameter not supported.',
        }


class IncludeInvalidPathError(Error):
    """Errors related to an invalid include relationship path."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, path):
        self.path = path

    def to_error(self):
        return {
            'status':
                '400',
            'source': {
                'parameter': 'include'
            },
            'title':
                'Invalid query parameter.',
            'detail': ("The resource does not have a "
                       f"'{self.path}' relationship path.")
        }


class FieldsInvalidResourceError(Error):
    """Errors related to an invalid fields resource."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, resource):
        self.resource = resource

    def to_error(self):
        return {
            'status': '400',
            'source': {
                'parameter': 'fields'
            },
            'title': 'Invalid query parameter.',
            'detail': f"Resource '{self.resource}' does not exist."
        }


class FieldsInvalidFieldError(Error):
    """Errors related to an invalid fields field."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, resource, field):
        self.resource = resource
        self.field = field

    def to_error(self):
        return {
            'status':
                '400',
            'source': {
                'parameter': 'fields'
            },
            'title':
                'Invalid query parameter.',
            'detail': (f"Field '{self.field}' does not exist in resource "
                       f"'{self.resource}'.")
        }


class FilterInvalidFieldError(Error):
    """Errors related to an invalid filter field."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, path):
        self.path = path

    def to_error(self):
        return {
            'status': '400',
            'source': {
                'parameter': 'filter'
            },
            'title': 'Invalid query parameter.',
            'detail': f"The resource does not have a '{self.path}' field."
        }


class SortInvalidFieldError(Error):
    """Errors related to an invalid sort field."""

    # pylint: disable=too-few-public-methods

    status = 400

    def __init__(self, path):
        self.path = path

    def to_error(self):
        return {
            'status': '400',
            'source': {
                'parameter': 'sort'
            },
            'title': 'Invalid query parameter.',
            'detail': f"The resource does not have a '{self.path}' field."
        }


class UnsupportedMediaTypeError(Error):
    """Error in response to a request with a Content-Type header that
    specifies the JSON API media type ('application/vnd.api+json') with any
    media type parameter.
    """

    # pylint: disable=too-few-public-methods

    status = 415

    def to_error(self):
        return {
            'status': '415',
            'source': {
                'header': 'Content-Type',
            },
            'title': 'Invalid header.',
        }


class NotAcceptableError(Error):
    """Error in response to a request with an Accept header that only have
    media type parameters specified when the media type in JSON API
    ('application/vnd.api+json').
    """

    # pylint: disable=too-few-public-methods

    status = 406

    def to_error(self):
        return {
            'status': '406',
            'source': {
                'header': 'Accept',
            },
            'title': 'Invalid header.',
        }


class DocumentNotJsonError(Error):
    """Error in response to a request with a document that is not JSON."""

    # pylint: disable=too-few-public-methods

    status = 400

    def to_error(self):
        return {
            'status': '400',
            'title': 'Invalid document.',
            'detail': 'Document is not JSON.',
        }


class DocumentDataMissingError(Error):
    """Error in response to a request with a document that is missing a
    top-level data member.
    """

    # pylint: disable=too-few-public-methods

    status = 422

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': '',
            },
            'title': 'Invalid document.',
            'detail': "Missing 'data' member at document's top level.",
        }


class DocumentError(Error, ABC):
    """Error in response to a request with a document error."""

    # pylint: disable=too-few-public-methods

    status = 422

    def __init__(self, pointer):
        super().__init__()
        self.pointer = pointer

    @abstractmethod
    def to_error(self):
        pass


class DocumentDataMalformedUuidError(DocumentError):
    """Error in response to a POST /{collection} request where the request's
    document contains a resource object with an id that is not a
    client-generated id (UUID).
    """

    # pylint: disable=too-few-public-methods

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Malformed client-generated id.',
        }


class DocumentDataInvalidAttributeError(DocumentError):
    """Error in response to a POST /{collection} request where the request's
    document contains a resource object with an invalid attribute.
    """

    # pylint: disable=too-few-public-methods

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Invalid attribute.',
        }


class DocumentDataInvalidRelationshipError(DocumentError):
    """Error in response to a POST /{collection} request where the request's
    document contains a resource object with an invalid relationship.
    """

    # pylint: disable=too-few-public-methods

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Invalid relationship.',
        }


class DocumentDataMissingTypeError(DocumentError):
    """Error in response to a POST /{collection} or a PATCH /{collection/{id}
    request where the request's document contains a resource object without a
    type.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, pointer='/data'):
        super().__init__(pointer)

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Missing type.',
        }


class DocumentDataMissingIdError(DocumentError):
    """Error in response to a PATCH /{collection/{id} request where the
    request document contains a resource object without an id.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, pointer='/data'):
        super().__init__(pointer)

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Missing id.',
        }


class DocumentDataNonexistentIdError(DocumentError):
    """Error in response to a PATCH
    /{collection}/{id}/relationships/{relationship} request where the request
    document contains a resource object with an id that does not exist.
    """

    # pylint: disable=too-few-public-methods

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource identifier object.',
            'detail': 'Nonexistent id.',
        }


class DocumentDataRelationshipConflictError(DocumentError):
    """Error in response to a PATCH
    /{collection}/{id}/relationships/{relationship} request where the request
    document contains a resource identifier object that is part of another
    one-to-one relationship already.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def __init__(self):
        super().__init__('/data')

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Conflicting resource identifier object.',
            'detail': 'Resource identifier object is part of another '
                      'one-to-one relationship.',
        }


class DocumentDataMalformedError(Error):
    """This error indicates that the request's document contains a malformed
    data member.
    """

    # pylint: disable=too-few-public-methods

    status = 422

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': '/data',
            },
            'title': 'Invalid resource object.',
            'detail': "Malformed 'data' member.",
        }


class DocumentDataInvalidTypeError(DocumentError):
    """Error in response to a POST /{collection} or PATCH /{collection}/{id}
    request where the request's document contains a resource object with an
    invalid type.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Invalid type.',
        }


class DocumentDataDuplicateIdError(DocumentError):
    """Error in response to a POST /{collection} request where the request's
    document contains an id that already exists so the resource object annot
    be created.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Id already exists.',
        }


class DocumentDataInvalidIdError(DocumentError):
    """Error in response to PATCH /{collection}/id request where the request
    document contains a resource object with an invalid id.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def __init__(self, title, pointer):
        super().__init__(pointer)
        self.title = title

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': self.title,
            'detail': 'Invalid id.',
        }


class DocumentDataMalformedIdError(DocumentError):
    """Error in response to PATCH /{collection}/id request where the request
    document contains a resource object with a malformed id.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def __init__(self, title, pointer):
        super().__init__(pointer)
        self.title = title

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': self.title,
            'detail': 'Malformed id.',
        }


class DocumentDataMissingRelationshipTypeError(DocumentError):
    """Error in response to a POST /{collection} or a PATCH /{collection/{id}
    request where the request's document contains a resource object with a
    relationship without a type.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, pointer='/data'):
        super().__init__(pointer)

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Missing relationship type.',
        }


class DocumentDataInvalidRelationshipTypeError(DocumentError):
    """Error in response to a POST /{collection} or PATCH /{collection}/{id}
    request where the request's document contains a resource object with a
    relationship with an invalid type.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Invalid relationship type.',
        }


class DocumentDataMissingRelationshipIdError(DocumentError):
    """Error in response to a POST /{collection} or PATCH /{collection/{id}
    request where the request document contains a resource object with a
    relationship without an id.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, pointer='/data'):
        super().__init__(pointer)

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Missing relationship id.',
        }


class DocumentDataMalformedRelationshipIdError(DocumentError):
    """Error in response to POST /{collection} or PATCH /{collection}/id
    request where the request document contains a resource object with a
    malformed id.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def __init__(self, title, pointer):
        super().__init__(pointer)
        self.title = title

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'pointer': self.pointer,
            },
            'title': self.title,
            'detail': 'Malformed relationship id.',
        }


class DocumentDataMalformedRelationshipError(DocumentError):
    """This error indicates that a request's document contains a malformed
    relationship.
    """

    # pylint: disable=too-few-public-methods

    status = 422

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Malformed relationship.',
        }


class DocumentDataMalformedRelationshipDataError(DocumentError):
    """This error indicates that a request's document contains a relationship
    whose data member is malformed.
    """

    # pylint: disable=too-few-public-methods

    status = 422

    def to_error(self):
        return {
            'status': '422',
            'source': {
                'pointer': self.pointer,
            },
            'title': 'Invalid resource object.',
            'detail': 'Malformed relationship data.',
        }


class DeleteObjectRemotelyRelatedError(Error):
    """This error indicates that the DELETE request of a resource object
    failed due to the resource object having an active remote relationship.
    """

    # pylint: disable=too-few-public-methods

    status = 409

    def __init__(self, path):
        self.path = path

    def to_error(self):
        return {
            'status': '409',
            'source': {
                'path': self.path,
            },
            'title': 'Cannot delete a remotely related resource object.',
        }


class AuthorizationError(Error):
    """This error indicates that the request did not contain a proper
    authentication bearer token.
    """

    # pylint: disable=too-few-public-methods

    status = 401

    def __init__(self, detail):
        self.detail = detail

    def to_error(self):
        return {
            'status': '401',
            'title': 'Unauthorized',
            'detail': self.detail,
        }



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module application provides class Application."""

from asyncio import sleep

from aiohttp import WSMsgType, web
from asyncpg.exceptions import (
    CannotConnectNowError,
    ConnectionDoesNotExistError,
)
from asyncpg.pool import create_pool

from ajsonapi.events import Event
from ajsonapi.exceptions import ErrorsException
from ajsonapi.init import init
from ajsonapi.requests import (
    parse_collection_document_query,
    parse_collection_query,
    parse_object_document_query,
    parse_object_query,
    parse_related_resource_query,
    parse_relationship_document_query,
    parse_relationship_query,
)
from ajsonapi.responses import errors_response
from ajsonapi.table import create_all as table_create_all
from ajsonapi.table import parse_model
from ajsonapi.uri.relationship import parse as parse_relationship


class Application:
    """JSON API web server.

    Default routes are derived from the specified object model.
    """

    def __init__(self, *args, **kwargs):
        Event.connect(self)
        self.app = web.Application(*args, **kwargs)
        self.ws_list = []
        self.pool = None

    async def connect_database(self, database):
        """Connects to the database.

        Args:
            database (str): URL to Postgres database. For example:
                'postgresql://user:password@localhost:5432/db'.

        Raises:
            NotImplementedError in case database doesn't start with
                'postgresql:'.
        """
        if not database.startswith('postgresql:'):
            raise NotImplementedError('Only postgres is currently supported.')
        while self.pool is None:
            try:
                self.pool = await create_pool(database)
            except (CannotConnectNowError,
                    ConnectionDoesNotExistError):  # pragma: no cover
                self.pool = None
                await sleep(.1)
        init(self.pool)

    async def disconnect_database(self):
        """Disconnects from the database."""
        await self.pool.close()

    def verify_model(self):
        """Parses the model."""
        # pylint: disable=no-self-use
        parse_model()

    async def create_tables(self):
        """Creates all tables in the database."""
        # pylint: disable=no-self-use
        await table_create_all(self.pool)

    def add_json_api_routes(self):
        """Add JSON API routes to the web application."""

        path = r'/{collection:[^/]*}'
        self.app.add_routes([
            web.get(path, self.get_collection),
            web.post(path, self.post_collection),
        ])

        path = r'/{collection:[^/]*}/{id:[0-9A-Fa-f-]*}'
        self.app.add_routes([
            web.get(path, self.get_object),
            web.delete(path, self.delete_object),
            web.patch(path, self.patch_object),
        ])

        path = (r'/{collection:[^/]*}/{id:[0-9A-Fa-f-]*}/'
                r'relationships/{relationship:[^/]*}')
        self.app.add_routes([
            web.get(path, self.get_relationship),
            web.patch(path, self.patch_relationship),
            web.post(path, self.post_relationship),
            web.delete(path, self.delete_relationship),
            web.put(path, self.put_relationship),
        ])

        path = (r'/{collection:[^/]*}/{id:[0-9A-Fa-f-]*}/'
                r'{related_resource:[^/]*}')
        self.app.add_routes([
            web.get(path, self.get_related_resource),
        ])

    def add_event_socket_route(self):
        """Add event socket."""

        path = '/event-socket/'
        self.app.add_routes([
            web.get(path, self.get_event_socket),
        ])

    async def get_collection(self, request):
        """Entry point for GET /{collection} requests.

        Args:
            request: The GET /{collection} request.
        """

        try:
            collection, query = parse_collection_query(request,
                                                       allow_include=True,
                                                       allow_fields=True,
                                                       allow_filter=True,
                                                       allow_sort=True,
                                                       allow_page=True)
            return await collection.get(query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def post_collection(self, request):
        """Entry point for GET /{collection} requests.

        Args:
            request: The POST /{collection} request.
        """

        try:
            collection, document, query = await parse_collection_document_query(
                request)
            return await collection.post(document, query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def get_object(self, request):
        """Entry point for GET /{collection}/{id} requests.

        Args:
            request: The GET /{collection}/{id} request.
        """
        try:
            object_, query = await parse_object_query(request,
                                                      allow_include=True,
                                                      allow_fields=True)
            return await object_.get(query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def delete_object(self, request):
        """Entry point for DELETE /{collection}/{id} requests.

        Args:
            request: The DELETE /{collection}/{id} request.
        """
        try:
            object_, _ = await parse_object_query(request)
            return await object_.delete()
        except ErrorsException as err:
            return errors_response(err.errors)

    async def patch_object(self, request):
        """Entry point for DELETE /{collection}/{id} requests.

        Args:
            request: The DELETE /{collection}/{id} request.
        """
        try:
            object_, document, query = await parse_object_document_query(request
                                                                        )
            return await object_.patch(document, query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def get_relationship(self, request):
        """Entry point for GET /{collection}/{id}/relationships/{relationship}
        requests.

        Args:
            request: The GET /{collection}/{id}/relationships/{relationship}
                request.

        Returns:
            Appropriate response for the request.
        """
        try:
            relationship, query = await parse_relationship_query(
                request, allow_include=True, allow_fields=True)
            return await relationship.get(query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def patch_relationship(self, request):
        """Entry point for PATCH
        /{collection}/{id}/relationships/{relationship} requests.

        Args:
            request: The PATCH /{collection}/{id}/relationships/{relationship}
                request.

        Returns:
            Appropriate response for the request.
        """
        try:
            relationship, document, query = \
                    await parse_relationship_document_query(request)
            return await relationship.patch(document, query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def post_relationship(self, request):
        """Entry point for POST
        /{collection}/{id}/relationships/{relationship} requests.

        Args:
            request: The POST /{collection}/{id}/relationships/{relationship}
                request.

        Returns:
            Appropriate response for the request.
        """
        try:
            relationship, document, query = \
                    await parse_relationship_document_query(request)
            return await relationship.post(document, query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def delete_relationship(self, request):
        """Entry point for DELETE
        /{collection}/{id}/relationships/{relationship} requests.

        Args:
            request: The DELETE /{collection}/{id}/relationships/{relationship}
                request.

        Returns:
            Appropriate response for the request.
        """
        try:
            relationship, document, query = \
                    await parse_relationship_document_query(request)
            return await relationship.delete(document, query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def put_relationship(self, request):
        """Entry point for PUT /{collection}/{id}/relationships/{relationship}
        requests.

        Args:
            request: The PUT /{collection}/{id}/relationships/{relationship}
                request.

        Returns:
            Appropriate response for the request.
        """
        relationship = parse_relationship(request)
        return relationship.put()

    async def get_related_resource(self, request):
        """Entry point for GET /{collection}/{id}/{related_resource} requests.

        Args:
            request: The GET /{collection}/{id}/{related_resource} request.

        Returns:
            Appropriate response for the request.
        """
        try:
            related, query = await parse_related_resource_query(
                request,
                allow_include=True,
                allow_fields=True,
                allow_filter=True,
                allow_sort=True,
                allow_page=True)
            return await related.get(query)
        except ErrorsException as err:
            return errors_response(err.errors)

    async def get_event_socket(self, request):
        """Creates a web socket for events."""

        ws = web.WebSocketResponse()  # pylint: disable=invalid-name
        await ws.prepare(request)
        self.ws_list.append(ws)
        async for msg in ws:
            if msg.type == WSMsgType.TEXT:
                if msg.data == 'close':
                    await ws.close()
                # Eventually: allow clients to register/deregister individual
                # events they are interested in.
            elif msg.type == WSMsgType.ERROR:
                print('websocket connection closed with exception',
                      ws.exception())
        print('websocket connection closed')
        self.ws_list.remove(ws)
        return ws

    async def broadcast(self, event):
        """Broadcasts an event on all event sockets."""

        stream_event = event.stream_event()
        for ws in self.ws_list:  # pylint: disable=invalid-name
            await ws.send_json(stream_event)



# Copyright © 2019-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module collection provides class GetCollectionIncludedEnvironment."""

from asyncio import Event, create_task, gather


class GetCollectionIncludedEnvironment:
    """Class GetCollectionIncludedEnvironment is the execution environment for
    the included part of a GET /{collection} response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, collection, ids, env):
        self.collection = collection
        self.ids = ids
        self.query = env.query
        self.tasks = []
        self.connection = env.connection
        self.lock = env.lock
        self.included_by_id = {}
        self.event = Event()

    def create_tasks(self):
        """Create all tasks for querying the database to create the included
        part of a GET /{collection} response.
        """
        self.tasks = [
            create_task(self.collection.get_collection_included(self, self.ids))
        ]
        query_fields = self.query['fields']
        if self.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_collection_included(self, self.ids))
                for rel in query_fields[self.collection]
                if rel in self.collection.table.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_collection_included(self, self.ids))
                for rel in self.collection.table.remote_relationships
            ])

    async def run(self):
        """Executes the GET /{collection} included tasks."""
        self.create_tasks()
        await gather(*self.tasks)


class GetObjectIncludedEnvironment:
    """Class GetObjectIncludedEnvironment is the execution environment for
    the included part of a GET /{collection}/{id} response.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, object_id, collection, ids, env):
        self.object_id = object_id
        self.collection = collection
        self.ids = ids
        self.query = env.query
        self.tasks = []
        self.connection = env.connection
        self.lock = env.lock
        self.included_by_id = {}
        self.event = Event()

    def create_tasks(self):
        """Create all tasks for querying the database to create the included
        part of a GET /{collection} response.
        """
        self.tasks = [
            create_task(self.collection.get_object_included(self, self.ids))
        ]
        query_fields = self.query['fields']
        if self.collection in query_fields:
            self.tasks.extend([
                create_task(rel.get_object_included(self, self.ids))
                for rel in query_fields[self.collection]
                if rel in self.collection.table.remote_relationships
            ])
        else:
            self.tasks.extend([
                create_task(rel.get_object_included(self, self.ids))
                for rel in self.collection.table.remote_relationships
            ])

    async def run(self):
        """Executes the GET /{collection} included tasks."""
        self.create_tasks()
        await gather(*self.tasks)



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module id provides class Id."""

from ajsonapi.types import Uuid


class Id:
    """Class representing the "id" column in JSON_API classes."""

    def __init__(self, type_):
        self.type_ = type_
        self.table_name = None  # Overriden in JSON_API.__init_subclass__

    def sql(self):
        """Returns the column definition for the "id" field."""
        # pylint: disable=no-self-use

        if self.type_ == Uuid:
            return "id UUID DEFAULT uuid_generate_v4()"
        # self.type_ == Int64
        return (
            f"id UUID DEFAULT CAST"
            f"(LPAD(TO_HEX(nextval('seq_{self.table_name}')), 32, '0') AS UUID)"
        )

    def sql_constraints(self):
        """Returns the constraints for the "id" field."""
        # pylint: disable=no-self-use
        return 'PRIMARY KEY (id)'

    def __str__(self):
        return 'id'



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
ajsonapi - asynchronous JSON API
================================

**ajsonapi** is a Python package providing the creation of a JSON API web
server backed by a database from a user-provided object model.
"""

from sys import version_info

from ajsonapi.application import Application
from ajsonapi.attribute import Attribute
from ajsonapi.errors import AuthorizationError
from ajsonapi.id import Id
from ajsonapi.json_api import JSON_API
from ajsonapi.relationships import (
    ManyToManyRelationship,
    ManyToOneRelationship,
    OneToManyRelationship,
    OneToOneLocalRelationship,
    OneToOneRemoteRelationship,
)
from ajsonapi.responses import errors_response
from ajsonapi.types import Boolean  # deprecating
from ajsonapi.types import DateTimeTimeZone  # deprecating
from ajsonapi.types import (
    Bool,
    DateTime,
    DateTimeTZ,
    Float32,
    Float64,
    Int16,
    Int32,
    Int64,
    Json,
    String,
    Uuid,
)

assert version_info >= (3, 7), 'Need Python 3.7+ for module ajsonapi'
del version_info

__docformat__ = 'restructuredtext'
__version__ = '0.7.1'



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module relationships provides relationship classes."""

from abc import abstractmethod

from asyncpg.exceptions import ForeignKeyViolationError

from ajsonapi.conversions import id_name_to_number, id_number_to_name
from ajsonapi.errors import (
    DocumentDataInvalidRelationshipTypeError,
    DocumentDataMalformedRelationshipDataError,
    DocumentDataMalformedRelationshipIdError,
    DocumentDataMissingRelationshipIdError,
    DocumentDataMissingRelationshipTypeError,
    DocumentDataNonexistentIdError,
)
from ajsonapi.exceptions import NonexistentRelationshipIdError
from ajsonapi.field import Field
from ajsonapi.id_value import IdValue
from ajsonapi.logging import log
from ajsonapi.uri.related import (
    ManyToManyRelatedResource,
    ManyToOneRelatedResource,
    OneToManyRelatedResource,
    OneToOneLocalRelatedResource,
    OneToOneRemoteRelatedResource,
)
from ajsonapi.uri.relationship import \
    ManyToManyRelationship as UriManyToManyRelationship
from ajsonapi.uri.relationship import \
    ManyToOneRelationship as UriManyToOneRelationship
from ajsonapi.uri.relationship import \
    OneToManyRelationship as UriOneToManyRelationship
from ajsonapi.uri.relationship import \
    OneToOneLocalRelationship as UriOneToOneLocalRelationship
from ajsonapi.uri.relationship import \
    OneToOneRemoteRelationship as UriOneToOneRemoteRelationship

# pylint: disable=too-many-lines


class Relationship(Field):
    """Abstract base class for all relationships between resources.
    """

    # pylint: disable=too-few-public-methods

    def __init__(self, rtable):
        self.rtable = rtable
        self.table = None  # Overridden in Table.__init_subclass__
        self.name = None  # Overridden in Table.__init_subclass__
        self.reverse = None  # Overridden in Table.__init_subclass__
        self.collection = None  # Overriden in Table.__init_subclass__

    @abstractmethod
    def make_uri_relationship(self, collection, id_):
        """Creates a URI Relationship object."""

    @abstractmethod
    def make_uri_related_resource(self, name, collection, id_):
        """Creates a URI Relationship object."""

    @abstractmethod
    async def verify_data_ids_exist(self, env, data_ids):
        """Verifies that a document's data/id members exist."""

    @abstractmethod
    def is_reverse(self, other):
        """Detects whether relationships are each other's reverse.

        Args:
            other: A Relationship instance.

        Returns:
           True if self and other are each other's reverse relationship, False
           otherwise.
        """

    @abstractmethod
    def collection_remote_ids(self, query, local_ids=None):
        """Returns the SQL statement for the remote ids of a relationship's
        local collection.
        """

    @abstractmethod
    def object_remote_ids(self, local_ids=None):
        """Returns the SQL condition for an ?include parameter from an
        object.
        """


class LocalRelationship(Relationship):
    """A relationship with a local foreign key."""

    def __init__(self, rtable, lfkey):
        super().__init__(rtable)
        self.lfkey = lfkey

    def filter_condition(self, values):
        uuids = ','.join(
            f"'{id_name_to_number(uuid)}'" for uuid in values.split(','))
        return f'{self.lfkey} IN ({uuids})'

    @abstractmethod
    def is_reverse(self, other):
        pass

    def sql(self):
        """Produces the SQL column definition for the local foreign key of
        this object.

        Returns:
            A string containing the SQL column definition for the local
            foreign key of this object.
        """

        return f'{self.lfkey} UUID'

    def __str__(self):
        return self.lfkey


class ToOneRelationship(Relationship):
    """A to-one relationship."""

    @abstractmethod
    def is_reverse(self, other):
        pass

    def verify_data_rios(self, data_relationship_name, data_relationship_data,
                         errors):
        """Verifies the resource identifier object specified in a to-one
        relationship in the request document.
        """
        relationship_type_name = self.rtable.collection.name
        data_rio = data_relationship_data

        # Relationship type
        try:
            data_rio_type_name = data_rio['type']
            if data_rio_type_name != relationship_type_name:
                errors.append(
                    DocumentDataInvalidRelationshipTypeError(
                        f'/data/relationships/{data_relationship_name}/data/'
                        f'type/{data_rio_type_name}'))
        except TypeError:
            errors.append(
                DocumentDataMalformedRelationshipDataError(
                    f'/data/relationships/{data_relationship_name}/data'))
            return None
        except KeyError:
            errors.append(
                DocumentDataMissingRelationshipTypeError(
                    f'/data/relationships/{data_relationship_name}/data'))

        # Relationship id
        try:
            data_rio_id_name = data_rio['id']
            try:
                return IdValue(data_rio_id_name)
            except ValueError:
                errors.append(
                    DocumentDataMalformedRelationshipIdError(
                        'Invalid resource identifier object.',
                        f'/data/relationships/{data_relationship_name}/data/'
                        f'id/{data_rio_id_name}'))
        except KeyError:
            errors.append(
                DocumentDataMissingRelationshipIdError(
                    f'/data/relationships/{data_relationship_name}/data'))

        return None

    async def verify_data_ids_exist(self, env, data_ids):
        data_id_number = data_ids.number
        stmt = f'SELECT FROM {self.rtable.name} WHERE id=$1'
        async with env.lock:
            log.debug(f'{stmt}: {data_id_number}')
            records = await env.connection.fetch(stmt, data_id_number)
        if records == []:
            data_id_name = data_ids.name
            env.errors.append(
                DocumentDataNonexistentIdError(
                    f'/data/relationships/{self.name}/data/id/{data_id_name}'))


class ToManyRelationship(Relationship):
    """A to-many relationship."""

    @abstractmethod
    def is_reverse(self, other):
        pass

    def verify_data_rios(self, data_relationship_name, data_relationship_data,
                         errors):
        """Verifies the resource identifier objects specified in a
        to-many relationship in the request document.
        """
        relationship_type_name = self.rtable.collection.name
        data_rio_ids = []
        for data_index, data_rio in enumerate(data_relationship_data):
            # Relationship type
            try:
                data_rio_type_name = data_rio['type']
                if data_rio_type_name != relationship_type_name:
                    errors.append(
                        DocumentDataInvalidRelationshipTypeError(
                            f'/data/relationships/{data_relationship_name}/'
                            f'data/{data_index}/type/{data_rio_type_name}'))
            except TypeError:
                errors.append(
                    DocumentDataMalformedRelationshipDataError(
                        f'/data/relationships/{data_relationship_name}/'
                        f'data/{data_index}'))
                continue
            except KeyError:
                errors.append(
                    DocumentDataMissingRelationshipTypeError(
                        f'/data/relationships/{data_relationship_name}/'
                        f'data/{data_index}'))

            # Relationship id
            try:
                data_rio_id_name = data_rio['id']
                try:
                    data_rio_id = IdValue(data_rio_id_name)
                    data_rio_ids.append(data_rio_id)
                except ValueError:
                    errors.append(
                        DocumentDataMalformedRelationshipIdError(
                            'Invalid resource identifier object.',
                            f'/data/relationships/{data_relationship_name}/'
                            f'data/{data_index}/id/{data_rio_id_name}'))
            except KeyError:
                errors.append(
                    DocumentDataMissingRelationshipIdError(
                        f'/data/relationships/{data_relationship_name}/'
                        f'data/{data_index}'))
        return data_rio_ids

    async def verify_data_ids_exist(self, env, data_ids):
        data_id_names = [data_id.name for data_id in data_ids]
        data_id_numbers = [data_id.number for data_id in data_ids]
        stmt = (f'SELECT id '
                f'FROM {self.rtable.name} '
                f'WHERE id=ANY($1::UUID[])')
        async with env.lock:
            log.debug(f'{stmt}: {data_id_numbers}')
            records = await env.connection.fetch(stmt, data_id_numbers)
        existing_id_names = {
            id_number_to_name(record['id']) for record in records
        }
        nonexistent_id_names = set(data_id_names) - existing_id_names
        env.errors.extend([
            DocumentDataNonexistentIdError(
                f'/data/relationships/{self.name}/data/id/{id_name}')
            for id_name in nonexistent_id_names
        ])


class OneToOneLocalRelationship(LocalRelationship, ToOneRelationship):
    """A one-to-one relationship between resources with a local foreign key.
    """

    def make_uri_relationship(self, collection, id_):
        return UriOneToOneLocalRelationship(collection, id_, self)

    def make_uri_related_resource(self, name, collection, id_):
        return OneToOneLocalRelatedResource(name, collection, id_, self)

    def is_reverse(self, other):
        return (isinstance(other, OneToOneRemoteRelationship) and
                other.table is self.rtable and other.rtable is self.table and
                other.rfkey == self.lfkey)

    def sql_constraints(self):
        """Returns the SQL unique and foreign key constraints for this
        OneToOneLocalRelationship.
        """

        return (f'UNIQUE ({self.lfkey}),\n   '
                f'FOREIGN KEY ({self.lfkey}) REFERENCES {self.rtable.name}(id)')

    def collection_remote_ids(self, query, local_ids=None):
        filters = query['filters']
        page = query['page']
        if local_ids is None and filters is None and page == {}:
            partial = ''
        else:
            sort = query['sort']
            if filters:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)}")
            else:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = f"SELECT id FROM {self.table.name}"
            if page:
                limit = page['size']
                offset = page['number'] * limit
                stmt += f' LIMIT {limit} OFFSET {offset}'
            partial = f' WHERE id IN ({stmt})'
        return f'SELECT {self.lfkey} FROM {self.table.name}{partial}'

    def object_remote_ids(self, local_ids=None):
        if local_ids is None:
            partial = '=$1'
        else:
            partial = f' IN ({local_ids})'
        return f'SELECT {self.lfkey} FROM {self.table.name} WHERE id{partial}'


class OneToOneRemoteRelationship(ToOneRelationship):
    """A one-to-one relationship between resources with a remote foreign key.
    """

    def __init__(self, rtable, rfkey):
        super().__init__(rtable)
        self.rfkey = rfkey

    def filter_condition(self, values):
        uuids = ','.join(
            f"'{id_name_to_number(uuid)}'" for uuid in values.split(','))
        return (f'id IN (SELECT {self.rfkey} from {self.rtable.name} '
                f'WHERE id IN ({uuids}))')

    def make_uri_relationship(self, collection, id_):
        return UriOneToOneRemoteRelationship(collection, id_, self)

    def make_uri_related_resource(self, name, collection, id_):
        return OneToOneRemoteRelatedResource(name, collection, id_, self)

    def is_reverse(self, other):
        return (isinstance(other, OneToOneLocalRelationship) and
                other.table is self.rtable and other.rtable is self.table and
                other.lfkey == self.rfkey)

    def collection_remote_ids(self, query, local_ids=None):
        filters = query['filters']
        page = query['page']
        if local_ids is None and filters is None and page == {}:
            partial = ' IS NOT NULL'
        else:
            sort = query['sort']
            if filters:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)}")
            else:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = f"SELECT id FROM {self.table.name}"
            if page:
                limit = page['size']
                offset = page['number'] * limit
                stmt += f' LIMIT {limit} OFFSET {offset}'
            partial = f' IN ({stmt})'
        return (f'SELECT id FROM {self.rtable.name} '
                f'WHERE {self.rfkey}{partial}')

    def object_remote_ids(self, local_ids=None):
        if local_ids is None:
            partial = '=$1'
        else:
            partial = f' IN ({local_ids})'
        return (f'SELECT id FROM {self.rtable.name} '
                f'WHERE {self.rfkey}{partial}')

    async def get_collection_data(self, env):
        """Creates the response data member for OneToOneRemoteRelationships.

        Specifically, after this call the environment's data_by_lid member is
        updated with something like the following:

        env.data_by_lid = {
            1: {
                'relationships': {
                    'one_one_remote': {
                        'data': {
                            'type': 'one_one_remotes',
                            'id': UUID_21,
                        },
                        'links': {
                            'self': '/centers/1/relationships/one_one_remote',
                            'related': '/centers/1/one_one_remote',
                        },
                    },
                },
            },
            2: {
                ...,
            },
            ...
        }
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IS NOT NULL')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
        for lid, data in env.data_by_lid.items():
            if lid in rid_by_lid:
                relationship_data = {
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid_by_lid[lid]),
                }
            else:
                relationship_data = None
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }

    async def get_collection_included(self, env, ids):
        """Gets the remote objects of this one-to-one remote relationship that
        need to be included in a GET /{collection} response.
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IN ({ids})')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
        for lid, data in env.included_by_id.items():
            if lid in rid_by_lid:
                relationship_data = {
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid_by_lid[lid])
                }
            else:
                relationship_data = None
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }

    async def post_collection_task(self, env):
        """Updates the database with the one-to-one (remote) relationship
        specified in a POST /{collection} request.
        """
        stmt = (f'UPDATE {self.rtable.name} '
                f'SET {self.rfkey} = $1 '
                f'WHERE id=$2')
        data_id = env.relationships[self.name]
        data_id_number = data_id.number
        await env.event.wait()
        async with env.lock:
            log.debug(f'{stmt}: {env.id_number}, {data_id_number}')
            result = await env.connection.execute(stmt, env.id_number,
                                                  data_id_number)
        if result == 'UPDATE 0':
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_object_data(self, env):
        """Creates the response data member for OneToOneRemoteRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_one_remote': {
                    'data': {
                        'type': 'one_one_remotes',
                        'id': UUID_21,
                    },
                    'links': {
                        'self': '/centers/1/relationships/one_one_remote',
                        'related': '/centers/1/one_one_remote',
                    },
                },
            },
        }
        """
        stmt = (f"SELECT id "
                f"FROM {self.rtable.name} "
                f"WHERE {self.rfkey} = $1")
        async with env.lock:
            log.debug(f'{stmt}: {env.obj.id.number}')
            record = await env.connection.fetchrow(stmt, env.obj.id.number)
        await env.event.wait()
        if record:
            rel_data = {
                'type': self.rtable.collection.name,
                'id': id_number_to_name(record['id']),
            }
        else:
            rel_data = None
        env.data.setdefault('relationships', {})[self.name] = {
            'data': rel_data,
            'links': {
                'self': (f'/{self.collection.name}/{env.obj.id.name}/'
                         f'relationships/{self.name}'),
                'related': (f'/{self.collection.name}/{env.obj.id.name}/'
                            f'{self.name}'),
            },
        }

    async def get_object_included(self, env, ids):
        """Gets the remote objects of this one-to-one remote relationship that
        need to be included in a GET response.
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IN ({ids})')
        async with env.lock:
            log.debug(f'{stmt}: {env.object_id}')
            records = await env.connection.fetch(stmt, env.object_id)
        await env.event.wait()
        rid_by_lid = {record[self.rfkey]: record['id'] for record in records}
        for lid, data in env.included_by_id.items():
            if lid in rid_by_lid:
                relationship_data = {
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid_by_lid[lid])
                }
            else:
                relationship_data = None
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                }
            }

    async def patch_object_task(self, env):
        """Updates the database with the one-to-one (remote) relationship
        specified in a POST /{collection} request.
        """
        stmt = (f'UPDATE {self.rtable.name} '
                f'SET {self.rfkey} = $1 '
                f'WHERE id=$2')
        data_id = env.relationships[self.name]
        data_id_number = data_id.number
        async with env.lock:
            log.debug(f'{stmt}: {env.obj.id.number}, {data_id_number}')
            result = await env.connection.execute(stmt, env.obj.id.number,
                                                  data_id_number)
        if result == 'UPDATE 0':
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_to_one_related_data(self, env):
        """Creates the response data member for OneToOneRemoteRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_one_remote': {
                    'data': {
                        'type': 'one_one_remotes',
                        'id': UUID_21,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    'links': {
                        'self': '/one_one_remotes/UUID_21',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, LocalRelationship):
            # OneToOneLocalRelationship or ManyToOneRelationship
            stmt = (f'SELECT id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT {env.related.relationship.lfkey}'
                    f' FROM {env.related.table.name}'
                    f' WHERE id = $1)')
        else:  # OneToOneRemoteRelationship
            stmt = (f'SELECT id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {env.related.id.number}')
            record = await env.connection.fetchrow(stmt, env.related.id.number)
        await env.event.wait()
        if env.data is None:
            return
        if record:
            rel_data = {
                'type': self.rtable.collection.name,
                'id': id_number_to_name(record[0]),
            }
        else:
            rel_data = None
        env.data.setdefault('relationships', {})[self.name] = {
            'data': rel_data,
            'links': {
                'self': (f'/{env.related.relationship.collection.name}/'
                         f'{env.related.id.name}')
            },
        }

    async def get_to_many_related_data(self, env):
        """Creates the response data member for OneToOneRemoteRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_one_remote': {
                    'data': {
                        'type': 'one_one_remotes',
                        'id': UUID_21,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    'links': {
                        'self': '/one_one_remotes/UUID_21',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, ManyToManyRelationship):
            stmt = (f'SELECT {self.rfkey}, id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT {env.related.relationship.rafkey}'
                    f' FROM {env.related.relationship.atable.name}'
                    f' WHERE {env.related.relationship.lafkey} = $1)')
        else:  # OneToManyRelationship
            stmt = (f'SELECT {self.rfkey}, id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {env.related.id.number}')
            records = await env.connection.fetch(stmt, env.related.id.number)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[0], []).append(record[1])
        for lid, data in env.data_by_lid.items():
            if lid in rids_by_lid:
                rids = rids_by_lid[lid]
                assert len(rids) == 1
                relationship_data = {
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rids[0]),
                }
            else:
                relationship_data = None
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{env.related.relationship.collection.name}/'
                             f'{env.related.id.name}')
                },
            }


class OneToManyRelationship(ToManyRelationship):
    """A one-to-many relationship between resources."""

    def __init__(self, rtable, rfkey):
        super().__init__(rtable)
        self.rfkey = rfkey

    def filter_condition(self, values):
        uuids = ','.join(
            f"'{id_name_to_number(uuid)}'" for uuid in values.split(','))
        return (f'id IN (SELECT {self.rfkey} FROM {self.rtable.name} '
                f'WHERE id in ({uuids}))')

    def make_uri_relationship(self, collection, id_):
        return UriOneToManyRelationship(collection, id_, self)

    def make_uri_related_resource(self, name, collection, id_):
        return OneToManyRelatedResource(name, collection, id_, self)

    def is_reverse(self, other):
        return (isinstance(other, ManyToOneRelationship) and
                other.table is self.rtable and other.rtable is self.table and
                other.lfkey == self.rfkey)

    def collection_remote_ids(self, query, local_ids=None):
        filters = query['filters']
        page = query['page']
        if local_ids is None and filters is None and page == {}:
            partial = ' IS NOT NULL'
        else:
            sort = query['sort']
            if filters:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)}")
            else:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = f"SELECT id FROM {self.table.name}"
            if page:
                limit = page['size']
                offset = page['number'] * limit
                stmt += f' LIMIT {limit} OFFSET {offset}'
            partial = f' IN ({stmt})'
        return (f'SELECT id FROM {self.rtable.name} '
                f'WHERE {self.rfkey}{partial}')

    def object_remote_ids(self, local_ids=None):
        if local_ids is None:
            partial = '=$1'
        else:
            partial = f' IN ({local_ids})'
        return (f'SELECT id FROM {self.rtable.name} '
                f'WHERE {self.rfkey}{partial}')

    async def get_collection_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data_by_lid member is
        updated with something like the following:

        env.data_by_lid = {
            1: {
                'relationships': {
                    'one_manys': {
                        'data': [
                            {
                                'type': 'one_manys',
                                'id': UUID_41,
                            },
                            ...
                        ],
                        'links': {
                            'self': '/centers/1/relationships/one_manys',
                            'related': '/centers/1/one_manys',
                        },
                    },
                },
            },
            2: {
                ...,
            },
            ...
        }
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IS NOT NULL')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.rfkey], []).append(record['id'])
        for lid, data in env.data_by_lid.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                }
            }

    async def get_collection_included(self, env, ids):
        """Gets the remote objects of this one-to-many relationship that need
        to be included in a GET /{collection} response.
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IN ({ids})')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.rfkey], []).append(record['id'])
        for lid, data in env.included_by_id.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                }
            }

    async def post_collection_task(self, env):
        """Updates the database with the one-to-many relationship specified in
        a POST /{collection} request.
        """
        stmt = (f'UPDATE {self.rtable.name} '
                f'SET {self.rfkey} = $1 '
                f'WHERE id=ANY($2::UUID[])')
        data_ids = env.relationships[self.name]
        data_id_numbers = [data_id.number for data_id in data_ids]
        await env.event.wait()
        async with env.lock:
            log.debug(f'{stmt}: {env.id_number}, {data_id_numbers}')
            result = await env.connection.execute(stmt, env.id_number,
                                                  data_id_numbers)
        if result == 'UPDATE 0':
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_object_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_manys': {
                    'data': [
                        {
                            'type': 'one_manys',
                            'id': UUID_41,
                        },
                        ...
                    ],
                    'links': {
                        'self': '/centers/1/relationships/one_manys',
                        'related': '/centers/1/one_manys',
                    },
                },
            },
        }
        """
        lfkey_name = self.rfkey
        stmt = (f"SELECT id "
                f"FROM {self.rtable.name} "
                f"WHERE {lfkey_name} = $1")
        async with env.lock:
            log.debug(f'{stmt}: {env.obj.id.number}')
            records = await env.connection.fetch(stmt, env.obj.id.number)
        await env.event.wait()
        rids = [record['id'] for record in records]
        env.data.setdefault('relationships', {})[self.name] = {
            'data': [{
                'type': self.rtable.collection.name,
                'id': id_number_to_name(rid),
            } for rid in rids],
            'links': {
                'self': (f'/{self.collection.name}/{env.obj.id.name}/'
                         f'relationships/{self.name}'),
                'related': (f'/{self.collection.name}/{env.obj.id.name}/'
                            f'{self.name}'),
            },
        }

    async def get_object_included(self, env, ids):
        """Gets the remote objects of this one-to-many relationship that need
        to be included in a GET response.
        """
        stmt = (f'SELECT {self.rfkey}, id '
                f'FROM {self.rtable.name} '
                f'WHERE {self.rfkey} IN ({ids})')
        async with env.lock:
            log.debug(f'{stmt}: {env.object_id}')
            records = await env.connection.fetch(stmt, env.object_id)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.rfkey], []).append(record['id'])
        for lid, data in env.included_by_id.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                }
            }

    async def patch_object_task(self, env):
        """Updates the database with the one-to-many relationship specified in
        a PATCH /{collection}/{id} request.
        """
        stmt = (f'UPDATE {self.rtable.name} '
                f'SET {self.rfkey} = $1 '
                f'WHERE id=ANY($2::UUID[])')
        data_ids = env.relationships[self.name]
        data_id_numbers = [data_id.number for data_id in data_ids]
        async with env.lock:
            log.debug(f'{stmt}: {env.obj.id.number}, {data_id_numbers}')
            result = await env.connection.execute(stmt, env.obj.id.number,
                                                  data_id_numbers)
        if result == 'UPDATE 0':
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_to_one_related_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_manys': {
                    'data': [{
                        'type': 'one_manys',
                        'id': UUID_41,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    ...
                    ],
                    'links': {
                        'self': '/one_manys/UUID_41',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, LocalRelationship):
            # OneToOneLocalRelationship or ManyToOneRelationship
            stmt = (f'SELECT id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT {env.related.relationship.lfkey}'
                    f' FROM {env.related.table.name}'
                    f' WHERE id = $1)')
        else:  # OneToOneRemoteRelationship
            stmt = (f'SELECT id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {env.related.id.number}')
            records = await env.connection.fetch(stmt, env.related.id.number)
        await env.event.wait()
        if env.data is None:
            return
        rel_data = [{
            'type': self.rtable.collection.name,
            'id': id_number_to_name(record[0])
        } for record in records]
        env.data.setdefault('relationships', {})[self.name] = {
            'data': rel_data,
            'links': {
                'self': f'/{self.rtable.collection.name}/{env.related.id.name}'
            },
        }

    async def get_to_many_related_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_manys': {
                    'data': [{
                        'type': 'one_manys',
                        'id': UUID_41,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    ...
                    ],
                    'links': {
                        'self': '/one_manys/UUID_41',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, ManyToManyRelationship):
            stmt = (f'SELECT {self.rfkey}, id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT {env.related.relationship.rafkey}'
                    f' FROM {env.related.relationship.atable.name}'
                    f' WHERE {env.related.relationship.lafkey} = $1)')
        else:  # OneToManyRelationship
            stmt = (f'SELECT {self.rfkey}, id '
                    f'FROM {self.rtable.name} '
                    f'WHERE {self.rfkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {env.related.id.number}')
            records = await env.connection.fetch(stmt, env.related.id.number)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[0], []).append(record[1])
        for lid, data in env.data_by_lid.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }


class ManyToOneRelationship(LocalRelationship, ToOneRelationship):
    """A many-to-one relationship between resources."""

    def make_uri_relationship(self, collection, id_):
        return UriManyToOneRelationship(collection, id_, self)

    def make_uri_related_resource(self, name, collection, id_):
        return ManyToOneRelatedResource(name, collection, id_, self)

    def is_reverse(self, other):
        return (isinstance(other, OneToManyRelationship) and
                other.rtable is self.table and other.rfkey == self.lfkey and
                other.rfkey == self.lfkey)

    def sql_constraints(self):
        """Returns the SQL foreign key constraint for this
        ManyToOneRelationship.
        """

        return f'FOREIGN KEY ({self.lfkey}) REFERENCES {self.rtable.name}(id)'

    def collection_remote_ids(self, query, local_ids=None):
        filters = query['filters']
        page = query['page']
        if local_ids is None and filters is None and page == {}:
            partial = ''
        else:
            sort = query['sort']
            if filters:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)}")
            else:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = f"SELECT id FROM {self.table.name}"
            if page:
                limit = page['size']
                offset = page['number'] * limit
                stmt += f' LIMIT {limit} OFFSET {offset}'
            partial = f' WHERE id IN ({stmt})'
        return f'SELECT {self.lfkey} FROM {self.table.name}{partial}'

    def object_remote_ids(self, local_ids=None):
        if local_ids is None:
            partial = '=$1'
        else:
            partial = f' IN ({local_ids})'
        return f'SELECT {self.lfkey} FROM {self.table.name} WHERE id{partial}'


class ManyToManyRelationship(ToManyRelationship):
    """A many-to-many relationship between resources."""

    def __init__(self, rtable, atable, lafkey, rafkey):
        # pylint: disable=too-many-arguments
        super().__init__(rtable)
        self.atable = atable
        self.lafkey = lafkey
        self.rafkey = rafkey

    def filter_condition(self, values):
        uuids = ','.join(
            f"'{id_name_to_number(uuid)}'" for uuid in values.split(','))
        return (f'id IN (SELECT {self.lafkey} FROM {self.atable.name} '
                f'WHERE {self.rafkey} IN ({uuids}))')

    def make_uri_relationship(self, collection, id_):
        return UriManyToManyRelationship(collection, id_, self)

    def make_uri_related_resource(self, name, collection, id_):
        return ManyToManyRelatedResource(name, collection, id_, self)

    def is_reverse(self, other):
        return (isinstance(other, ManyToManyRelationship) and
                other.table is self.rtable and other.rtable is self.table and
                other.atable is self.atable and other.lafkey == self.rafkey and
                other.rafkey == self.lafkey)

    def collection_remote_ids(self, query, local_ids=None):
        filters = query['filters']
        page = query['page']
        if local_ids is None and filters is None and page == {}:
            partial = ' IS NOT NULL'
        else:
            sort = query['sort']
            if filters:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"WHERE {' AND '.join(filters)}")
            else:
                if sort:
                    stmt = (f"SELECT id FROM {self.table.name} "
                            f"ORDER BY {', '.join(sort)}")
                else:
                    stmt = f"SELECT id FROM {self.table.name}"
            if page:
                limit = page['size']
                offset = page['number'] * limit
                stmt += f' LIMIT {limit} OFFSET {offset}'
            partial = f' IN ({stmt})'
        return (f'SELECT {self.rafkey} FROM {self.atable.name} '
                f'WHERE {self.lafkey}{partial}')

    def object_remote_ids(self, local_ids=None):
        if local_ids is None:
            partial = '=$1'
        else:
            partial = f' IN ({local_ids})'
        return (f'SELECT {self.rafkey} FROM {self.atable.name} '
                f'WHERE {self.lafkey}{partial}')

    async def get_collection_data(self, env):
        """Creates the response data member for ManyToManyRelationships.

        Specifically, after this call the environment's data_by_lid member is
        updated with something like the following:

        env.data_by_lid = {
            1: {
                'relationships': {
                    'many_manys': {
                        'data': [
                            {
                                'type': 'many_manys',
                                'id': UUID_51,
                            },
                            ...
                        ],
                        'links': {
                            'self': '/centers/1/relationships/many_manys',
                            'related': '/centers/1/many_manys',
                        },
                    },
                },
            },
            2: {
                ...,
            },
            ...
        }
        """
        stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                f'FROM {self.atable.name}')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.lafkey],
                                   []).append(record[self.rafkey])
        for lid, data in env.data_by_lid.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }

    async def get_collection_included(self, env, ids):
        """Gets the remote objects of this many-to-many relationship that need
        to be included in a GET /{collection} response.
        """
        stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                f'FROM {self.atable.name} '
                f'WHERE {self.lafkey} IN ({ids})')
        async with env.lock:
            log.debug(stmt)
            records = await env.connection.fetch(stmt)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.lafkey],
                                   []).append(record[self.rafkey])
        for lid, data in env.included_by_id.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }

    async def post_collection_task(self, env):
        """Updates the database with the many-to-many relationship specified
        in a POST /{collection} request.
        """
        stmt = (f'INSERT INTO {self.atable.name} '
                f'({self.lafkey}, {self.rafkey}) '
                f'VALUES ($1, $2)')
        data_ids = env.relationships[self.name]
        await env.event.wait()
        data = [(env.id_number, data_id.number) for data_id in data_ids]
        try:
            async with env.lock:
                async with env.connection.transaction():
                    log.debug(f'{stmt}: {data}')
                    await env.connection.executemany(stmt, data)
        except ForeignKeyViolationError:
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_object_data(self, env):
        """Creates the response data member for ManyToManyRelationships.

        Specifically, after this call the environment's data is updated with
        something like the following:

        env.data = {
            'relationships': {
                'many_manys': {
                    'data': [
                        {
                            'type': 'many_manys',
                            'id': UUID_51,
                        },
                        ...
                    ],
                    'links': {
                        'self': '/centers/1/relationships/many_manys',
                        'related': '/centers/1/many_manys',
                    },
                },
            }
        }
        """
        stmt = (f"SELECT {self.rafkey} "
                f"FROM {self.atable.name} "
                f"WHERE {self.lafkey} = $1")
        async with env.lock:
            records = await env.connection.fetch(stmt, env.obj.id.number)
        await env.event.wait()
        rids = [record[self.rafkey] for record in records]
        rafkeys_by_lafkey = {}
        for record in records:
            rafkeys_by_lafkey.setdefault(env.obj.id.number,
                                         []).append(record[self.rafkey])
        env.data.setdefault('relationships', {})[self.name] = {
            'data': [{
                'type': self.rtable.collection.name,
                'id': id_number_to_name(rid),
            } for rid in rids],
            'links': {
                'self': (f'/{self.collection.name}/{env.obj.id.name}/'
                         f'relationships/{self.name}'),
                'related': (f'/{self.collection.name}/{env.obj.id.name}/'
                            f'{self.name}'),
            },
        }

    async def get_object_included(self, env, ids):
        """Gets the remote objects of this many-to-many relationship that need
        to be included in a GET response.
        """
        stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                f'FROM {self.atable.name} '
                f'WHERE {self.lafkey} IN ({ids})')
        async with env.lock:
            records = await env.connection.fetch(stmt, env.object_id)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[self.lafkey],
                                   []).append(record[self.rafkey])
        for lid, data in env.included_by_id.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            lid_name = id_number_to_name(lid)
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self': (f'/{self.collection.name}/{lid_name}/'
                             f'relationships/{self.name}'),
                    'related': (f'/{self.collection.name}/{lid_name}/'
                                f'{self.name}'),
                },
            }

    async def patch_object_task(self, env):
        """Updates the database with the many-to-many relationship specified
        in a PATCH /{collection}/{id} request.
        """
        stmt = (f'INSERT INTO {self.atable.name} '
                f'({self.lafkey}, {self.rafkey}) '
                f'VALUES ($1, $2)')
        data_ids = env.relationships[self.name]
        data = [(env.obj.id.number, data_id.number) for data_id in data_ids]
        try:
            async with env.lock:
                async with env.connection.transaction():
                    log.debug(f'{stmt}: {data}')
                    await env.connection.executemany(stmt, data)
        except ForeignKeyViolationError:
            env.cancel()
            raise NonexistentRelationshipIdError()

    async def get_to_one_related_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_manys': {
                    'data': [{
                        'type': 'one_manys',
                        'id': UUID_41,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    ...
                    ],
                    'links': {
                        'self': '/one_manys/UUID_41',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, LocalRelationship):
            # OneToOneLocalRelationship or ManyToOneRelationship
            stmt = (f'SELECT {self.rafkey} '
                    f'FROM {self.atable.name} '
                    f'WHERE {self.lafkey} IN ('
                    f' SELECT {env.related.relationship.lfkey}'
                    f' FROM {env.related.table.name}'
                    f' WHERE id = $1)')
        else:  # OneToOneRemoteRelationship
            stmt = (f'SELECT {self.rafkey} '
                    f'FROM {self.atable.name} '
                    f'WHERE {self.lafkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}, {env.related.id.number}')
            records = await env.connection.fetch(stmt, env.related.id.number)
        await env.event.wait()
        if env.data is None:
            return
        rids = [record[0] for record in records]
        relationship_data = [{
            'type': self.rtable.collection.name,
            'id': id_number_to_name(rid),
        } for rid in rids]
        env.data.setdefault('relationships', {})[self.name] = {
            'data': relationship_data,
            'links': {
                'self': f'/{env.related.collection.name}/{env.related.id.name}'
            },
        }

    async def get_to_many_related_data(self, env):
        """Creates the response data member for OneToManyRelationships.

        Specifically, after this call the environment's data member is updated
        with something like the following:

        env.data = {
            'relationships': {
                'one_manys': {
                    'data': [{
                        'type': 'one_manys',
                        'id': UUID_41,
                        'attributes': {
                            ...
                        },
                        'relationships': {
                            ...
                        },
                    },
                    ...
                    ],
                    'links': {
                        'self': '/one_manys/UUID_41',
                    },
                },
            },
        }
        """
        if isinstance(env.related.relationship, ManyToManyRelationship):
            stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                    f'FROM {self.atable.name} '
                    f'WHERE {self.lafkey} IN ('
                    f' SELECT {env.related.relationship.rafkey}'
                    f' FROM {env.related.relationship.atable.name}'
                    f' WHERE {env.related.relationship.lafkey} = $1)')
        else:  # OneToManyRelationship
            stmt = (f'SELECT {self.lafkey}, {self.rafkey} '
                    f'FROM {self.atable.name} '
                    f'WHERE {self.lafkey} IN ('
                    f' SELECT id'
                    f' FROM {env.related.relationship.rtable.name}'
                    f' WHERE {env.related.relationship.rfkey} = $1)')
        async with env.lock:
            log.debug(f'{stmt}: {env.related.id.number}')
            records = await env.connection.fetch(stmt, env.related.id.number)
        await env.event.wait()
        rids_by_lid = {}
        for record in records:
            rids_by_lid.setdefault(record[0], []).append(record[1])
        for lid, data in env.data_by_lid.items():
            if lid in rids_by_lid:
                relationship_data = [{
                    'type': self.rtable.collection.name,
                    'id': id_number_to_name(rid),
                } for rid in rids_by_lid[lid]]
            else:
                relationship_data = []
            data.setdefault('relationships', {})[self.name] = {
                'data': relationship_data,
                'links': {
                    'self':
                        f'/{env.related.collection.name}/{env.related.id.name}'
                },
            }



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module query deals with everything related to query parameters."""

from re import compile as re_compile

from ajsonapi.errors import (
    FieldsInvalidFieldError,
    FieldsInvalidResourceError,
    QueryParameterUnsupportedError,
)
from ajsonapi.field import Field

RE_FIELDS = re_compile(r'fields\[([^\]]*)\]')
RE_PAGE = re_compile(r'page\[([^\]]*)\]')


def query_has_parameter(query, name):
    """Checks if a query parameter exists."""
    for parameter in query:
        if parameter.split('[')[0] == name:
            return True
    return False


def parse_fields(request_query, query, errors, *, allow=False):
    """Parses the request's fields query parameter."""

    if query_has_parameter(request_query, 'fields'):
        if allow:
            query['fields'] = fields_dict(request_query, errors)
        else:
            errors.append(QueryParameterUnsupportedError('fields'))
    else:
        query['fields'] = {}


def fields_dict(request_query, errors):
    """Helper function for parsing request query fields parameters."""

    # pylint: disable=import-outside-toplevel,cyclic-import
    from ajsonapi.uri.collection import Collection

    return_dict = {}
    for key, values in request_query.items():
        match = RE_FIELDS.match(key)
        if match:
            collection_name = match.group(1)
            try:
                collection = Collection.by_name[collection_name]
            except KeyError:
                errors.append(FieldsInvalidResourceError(collection_name))
                continue
            return_dict[collection] = []
            if values == '':
                continue
            for field_name in values.split(','):
                try:
                    field = getattr(collection.table.___, field_name)
                except AttributeError:
                    errors.append(
                        FieldsInvalidFieldError(collection_name, field_name))
                    continue
                if not isinstance(field, Field):
                    errors.append(
                        FieldsInvalidFieldError(collection_name, field_name))
                else:
                    return_dict[collection].append(field)
    return return_dict


def parse_page(request_query, query, errors, *, allow=False):
    """Parses the request's page query parameter."""
    if query_has_parameter(request_query, 'page'):
        if allow:
            query['page'] = page_dict(request_query, errors)
        else:
            errors.append(QueryParameterUnsupportedError('page'))
    else:
        query['page'] = {}


def page_dict(request_query, errors):
    """Helper function for parsing request query page parameters."""
    # pylint: disable=unused-argument

    return_dict = {}
    for key, value in request_query.items():
        match = RE_PAGE.match(key)
        if match:
            page_key = match.group(1)
            if page_key == 'number':
                return_dict['number'] = int(value)
            elif page_key == 'size':
                return_dict['size'] = int(value)
    return return_dict



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module requests deals with everything related to requests."""

from ajsonapi.headers import parse as parse_headers
from ajsonapi.uri.collection import parse_collection
from ajsonapi.uri.object import parse as parse_object
from ajsonapi.uri.related import parse as parse_related_resource
from ajsonapi.uri.relationship import parse as parse_relationship


def parse_collection_query(request,
                           allow_include=False,
                           allow_fields=False,
                           allow_filter=False,
                           allow_sort=False,
                           allow_page=False):
    """Gets the resource collection and query parameters from the request."""
    # pylint: disable=too-many-arguments

    parse_headers(request)
    collection = parse_collection(request)
    query = collection.parse_query(request,
                                   allow_include=allow_include,
                                   allow_fields=allow_fields,
                                   allow_filter=allow_filter,
                                   allow_sort=allow_sort,
                                   allow_page=allow_page)
    return collection, query


async def parse_collection_document_query(request, allow_include=False):
    """Gets the resource collection, the data member from the document, and
    query parameters from the request.
    """

    parse_headers(request)
    collection = parse_collection(request)
    query = collection.parse_query(request, allow_include=allow_include)
    document = await collection.parse_document(request)
    return collection, document, query


async def parse_object_query(request, allow_include=False, allow_fields=False):
    """Gets the resource object and query parameters from the request."""

    parse_headers(request)
    object_ = parse_object(request)
    query = object_.collection.parse_query(request,
                                           allow_include=allow_include,
                                           allow_fields=allow_fields)
    return object_, query


async def parse_object_document_query(request):
    """Gets the resource object, the data member from the document, and query
    parameters from the request.
    """

    parse_headers(request)
    object_ = parse_object(request)
    query = object_.collection.parse_query(request)
    document = await object_.parse_document(request)
    return object_, document, query


async def parse_relationship_query(request,
                                   allow_include=False,
                                   allow_fields=False):
    """Gets the relationship and query parameters from the request."""

    parse_headers(request)
    relationship = parse_relationship(request)
    query = relationship.collection.parse_query(request,
                                                allow_include=allow_include,
                                                allow_fields=allow_fields)
    return relationship, query


async def parse_relationship_document_query(request, allow_include=False):
    """Gets the relationship and query parameters from the request."""

    parse_headers(request)
    relationship = parse_relationship(request)
    query = relationship.collection.parse_query(request,
                                                allow_include=allow_include)
    document = await relationship.parse_document(request)
    return relationship, document, query


async def parse_related_resource_query(request,
                                       allow_include=False,
                                       allow_fields=False,
                                       allow_filter=False,
                                       allow_sort=False,
                                       allow_page=False):
    """Gets the related resource and query parameters from the request."""
    # pylint: disable=too-many-arguments

    parse_headers(request)
    related_resource = parse_related_resource(request)
    query = related_resource.relationship.rtable.collection.parse_query(
        request,
        allow_include=allow_include,
        allow_fields=allow_fields,
        allow_filter=(allow_filter and related_resource.is_to_many()),
        allow_sort=(allow_sort and related_resource.is_to_many()),
        allow_page=(allow_page and related_resource.is_to_many()))
    return related_resource, query



# Copyright © 2018-2020 Roel van der Goot
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Module table provides class Table."""

from itertools import chain

from ajsonapi.attribute import Attribute
from ajsonapi.conversions import pascal_to_snake
from ajsonapi.exceptions import ModelError
from ajsonapi.logging import log
# pylint: disable=unused-import
from ajsonapi.relationships import (
    LocalRelationship,
    ManyToManyRelationship,
    ManyToOneRelationship,
    Relationship,
)
from ajsonapi.toposort import CircularDependencyError, toposort
from ajsonapi.types import Int64
from ajsonapi.uri.collection import Collection


class Original:
    """Class to store the original Table attributes and relationships."""

    # pylint: disable=too-few-public-methods

    def _pylint(self):
        """Method to stop pylint from complaining about too-few-public-methods
        in an empty class with a disable=too-few-public-methods directive
        (PyCQA/pylint issue #2913).
        """


class Table:
    """Class Table is the Python representation of an SQL table."""
    # pylint: disable=too-few-public-methods

    by_class_name = {}

    @classmethod
    def __init_subclass__(cls):
        # pylint: disable=too-many-statements,too-many-branches,too-many-nested-blocks
        if cls.__name__ in ['JSON_API', 'AssociationTable']:
            return

        # To prevent ajsonapi's attributes and relationships from overriding the
        # user's attributes and relationships, move the user's attributes and
        # relationships to ___.
        cls.___ = Original()
        for name, col in cls.__dict__.items():
            if isinstance(col, (Attribute, Relationship)):
                setattr(cls.___, name, col)
        for name, col in cls.___.__dict__.items():
            if isinstance(col, (Attribute, Relationship)):
                delattr(cls, name)

        cls.name = cls.__name__
        collection_name = pascal_to_snake(cls.__name__)
        cls.collection = Collection(collection_name, cls)
        Collection.by_name[collection_name] = cls.collection
        cls.attributes = []
        cls.attribute_names = []
        cls.relationships = []
        cls.local_relationships = []
        cls.remote_relationships = []
        cls.lfkey_by_relationship_name = {}
        cls.relationship_by_lfkey = {}
        cls.columns = []
        cls.constraints = []
        cls.pool = None

        # Update relationships to class 'cls' for classes in cls.by_class_name.
        for table in Table.by_class_name.values():
            for col in table.___.__dict__.values():
                if isinstance(col, Relationship):
                    if col.rtable == cls.__name__:
                        col.rtable = cls
                    if isinstance(col, ManyToManyRelationship):
                        if col.atable == cls.__name__:
                            col.atable = cls

        # Now insert cls in by_class_name in case it contains a relationship to
        # itself. That way the next code section resolves the circular
        # dependency.
        Table.by_class_name[cls.name] = cls

        # Update attributes and relationships in cls.
        for name, col in cls.___.__dict__.items():
            if isinstance(col, Attribute):
                col.name = name
                cls.columns.append(col)
                cls.attributes.append(col)
                cls.attribute_names.append(col.name)
            elif isinstance(col, Relationship):
                col.name = name
                cls.relationships.append(col)
                col.table = cls
                col.collection = cls.collection
                if isinstance(col, LocalRelationship):
                    cls.local_relationships.append(col)
                    cls.lfkey_by_relationship_name[col.name] = col.lfkey
                    cls.relationship_by_lfkey[col.lfkey] = col
                    cls.columns.append(col)
                    cls.constraints.append(col)
                else:
                    cls.remote_relationships.append(col)
                    if (isinstance(col, ManyToManyRelationship) and
                            isinstance(col.atable, str)):
                        if col.atable in Table.by_class_name:
                            col.atable = Table.by_class_name[col.atable]
                        else:
                            lclassname = cls.__name__
                            if isinstance(col.rtable, str):
                                rclassname = col.rtable
                            else:
                                rclassname = col.rtable.__name__
                            cls_def = (
                                f"class {col.atable}(AssociationTable):\n"
                                f"    rel0 = ManyToOneRelationship"
                                f"('{lclassname}', lfkey='{col.lafkey}')\n"
                                f"    rel1 = ManyToOneRelationship"
                                f"('{rclassname}', lfkey='{col.rafkey}')\n")
                            # pylint: disable=exec-used
                            exec(compile(cls_def, '<string>', 'exec'))
                if (isinstance(col.rtable, str) and
                        col.rtable in Table.by_class_name):
                    col.rtable = Table.by_class_name[col.rtable]
                    for rel in col.rtable.relationships:
                        if col.is_reverse(rel):
                            rel.reverse = col
                            col.reverse = rel
                            break

    @classmethod
    def sql_statements(cls, exclude_classes=None):
        """Creates the sql statements for this table.

        Returns:
            stmt_table: The SQL statement to create the table.
            stmts_fk: List of SQL statements for foreign key constraints that
                are removed from the stmt_table.
        """
        if exclude_classes is None:
            exclude_classes = set()
        columns_now, columns_later = [], []
        for col in cls.columns:
            if col in cls.local_relationships:
                if col.rtable in exclude_classes:
                    columns_later.append(col)
                else:
                    columns_now.append(col)
            else:
                columns_now.append(col)
        table_fields = ',\n   '.join(
            chain((col.sql() for col in cls.columns),
                  (col.sql_constraints()
                   for col in cls.constraints
                   if col in columns_now)))
        # pylint: disable=no-member
        if hasattr(cls, 'id') and cls.id.type_ == Int64:
            stmt_table = (f'CREATE SEQUENCE IF NOT EXISTS seq_{cls.name};\n'
                          f'CREATE TABLE IF NOT EXISTS {cls.name} (\n   '
                          f'{table_fields}\n);')
        else:
            stmt_table = (f'CREATE TABLE IF NOT EXISTS {cls.name} (\n   '
                          f'{table_fields}\n);')
        stmts_fk = list(f'ALTER TABLE {cls.name} ADD {col.sql_constraints()}'
                        for col in columns_later)
        return stmt_table, stmts_fk

    @classmethod
    def _dependencies(cls):
        """Fetches the creation dependencies for this class.

        Returns:
            A set of classes whose tables must be created before this class's
            table.
        """
        return {rel.rtable for rel in cls.local_relationships}


class AssociationTable(Table):
    """Class AssociationTable is the Python representation of an SQL
    association table.
    """
    # pylint: disable=too-few-public-methods

    @classmethod
    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.constraints.append(cls)

    @classmethod
    def sql_constraints(cls):
        """Returns the constraints for classes derived from AssociationTable."""
        return ("UNIQUE "
                f"({', '.join(rel.lfkey for rel in cls.local_relationships)})")


def parse_model():
    """Parse the provided model for failures."""

    errors = []
    for class_name, table in Table.by_class_name.items():
        for rel in table.relationships:
            if isinstance(rel.table, str):
                error = (f'Relationship {class_name}.{rel.name}: '
                         f'remote table {rel.table} does not exist.')
                errors.append(error)
    if errors:
        raise ModelError(errors)


async def db_execute(pool, stmt):
    """Execute statement."""

    async with pool.acquire() as connection:
        log.debug(stmt)
        return await connection.execute(stmt)


async def create_all(pool):
    """Creates the SQL tables for the classes in the object model.

    Args:
        pool: Connection pool to the database in which to create the SQL tables.
    """

    await db_execute(pool, 'CREATE EXTENSION IF NOT EXISTS "uuid-ossp";')

    # pylint: disable=protected-access
    stmts_delayed = []  # Delayed foreign key constraints
    dependencies = {
        table: table._dependencies() for table in Table.by_class_name.values()
    }
    while True:
        try:
            for tables in toposort(dependencies):
                for table in tables:
                    stmt_table, _ = table.sql_statements()
                    await db_execute(pool, stmt_table)
            break
        except CircularDependencyError as exc:
            dependencies = exc.data
            itr = iter(dependencies.items())
            table, deps = next(itr)
            stmt_table, stmts_fk = table.sql_statements(deps)
            await db_execute(pool, stmt_table)
            stmts_delayed.extend(stmts_fk)
            dependencies = {
                item: (dep - {table})
                for item, dep in dependencies.items()
                if item != table
            }
    for stmt in stmts_delayed:
        await db_execute(pool, stmt)


def init(pool):
    """Initializes all user Table subclasses."""

    for table in Table.__subclasses__():
        if table.__name__ != 'JSON_API':
            table.pool = pool
    for table in AssociationTable.__subclasses__():
        table.pool = pool
